using System;
using System.Collections.Generic;
using System.Text;
using EnvDTE;
using Simplabs.NetDocumentor.Documentation;
using Simplabs.NetDocumentor.Settings;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Diagnostics;
using EnvDTE80;
using Microsoft.VisualStudio.VCCodeModel;
using System.IO;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// Documentor for Basic code
	/// </summary>
	internal class BasicDocumentor : ManagedCodeDocumentor, IDocumentor {

		/// <summary>
		/// The language prefix
		/// </summary>
		/// <remarks>This prefix is used to retrieve confiiguration settings etc. for this language</remarks>
		internal const string LanguagePrefix = "Basic";

		/// <summary>
		/// The file extensions the documentor supports
		/// </summary>
		private string[] extensions = { ".vb" };

		/// <summary>
		/// Specifies whether the documentor suppresses the UI
		/// </summary>
		private bool suppressUI;

		/// <summary>
		/// Initializes a new BasicDocumentor
		/// </summary>
		internal BasicDocumentor() {
		}

		/// <summary>
		/// Specifies whether the documentor should suppress the UI
		/// </summary>
		bool IDocumentor.SuppressUI {
			get {
				return suppressUI;
			}
			set {
				suppressUI = value;
			}
		}

		/// <summary>
		/// Event that is fired whenever a code element is documented
		/// </summary>
		public event CodeElementDocumentedHandler ElementDocumented;

		/// <summary>
		/// Applies the documentation comment <paramref name="comment"/> to the specified code element
		/// </summary>
		/// <param name="element">The code element to apply the documentation comment <paramref name="comment"/> to</param>
		/// <param name="comment">The comment to format</param>
		/// <returns>A formatted string that is assignable to code elements</returns>
		private void ApplyDocumentation(CodeElement2 element, DocumentationComment comment) {
			SourceCodeServices.InsertDocumentation(element, comment.ToString());
		}

		/// <summary>
		/// Initializes the documentor
		/// </summary>
		void IDocumentor.Initialize() {
			suppressUI = false;
		}

		/// <summary>
		/// Generates the XML documentation structure for a given code <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to document</param>
		/// <param name="documentRecursively">Set to true to document <paramref name="element"/> and alll of its members recursively or set to false to document just <paramref name="element"/></param>
		void IDocumentor.DocumentElement(CodeElement2 element, bool documentRecursively) {
			DocumentElement(element, documentRecursively, false);
		}

		/// <summary>
		/// Generates the XML documentation structure for a given code <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to document</param>
		/// <param name="documentRecursively">Set to true to document <paramref name="element"/> and alll of its members recursively or set to false to document just <paramref name="element"/></param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="elemment"/>. This is used if existing comments are set to be retained and <paramref name="element"/> itself has a comment. In this case, the mebers of <paramref name="element"/> still could have no comments.</param>
		private void DocumentElement(CodeElement2 element, bool documentRecursively, bool membersOnly) {
			try {
				bool preserveComment = true;
				if (documentRecursively && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
					preserveComment = false;
				}
				DocumentationComment comment = null;
				switch (element.Kind) {
					case vsCMElement.vsCMElementClass:
						if (preserveComment) {
							comment = DocumentClass(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentClass(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementFunction:
						if (!membersOnly) {
							comment = DocumentFunction(element);
						}
						break;
					case vsCMElement.vsCMElementProperty:
						if (!membersOnly) {
							comment = DocumentProperty(element);
						}
						break;
					case vsCMElement.vsCMElementVariable:
						if (!membersOnly) {
							comment = DocumentField(element);
						}
						break;
					case vsCMElement.vsCMElementEvent:
						if (!membersOnly) {
							comment = DocumentEvent(element);
						}
						break;
					case vsCMElement.vsCMElementInterface:
						if (preserveComment) {
							comment = DocumentInterface(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentInterface(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementStruct:
						if (preserveComment) {
							comment = DocumentStruct(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentStruct(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementEnum:
						if (preserveComment) {
							comment = DocumentEnum(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentEnum(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementDelegate:
						if (!membersOnly) {
							comment = DocumentDelegate(element);
						}
						break;
					case vsCMElement.vsCMElementModule:
						if (!membersOnly) {
							comment = DocumentModule(element);
						}
						break;
					case vsCMElement.vsCMElementNamespace:
						if (documentRecursively) {
							comment = DocumentNamespace(element, true);
						}
						break;
					default:
						return;
				}
				if (ElementDocumented != null) {
					ElementDocumented(element, comment != null ? comment.IsIncomplete : false);
				}
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> will be documented according to the current settings
		/// </summary>
		/// <param name="element">The element to check for whether it will be documented or not</param>
		bool IDocumentor.GetElementWillBeDocumented(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
				case vsCMElement.vsCMElementFunction:
				case vsCMElement.vsCMElementProperty:
				case vsCMElement.vsCMElementVariable:
				case vsCMElement.vsCMElementEvent:
				case vsCMElement.vsCMElementInterface:
				case vsCMElement.vsCMElementStruct:
				case vsCMElement.vsCMElementEnum:
				case vsCMElement.vsCMElementModule:
					return true;
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> can be recursively documented
		/// </summary>
		/// <param name="element">The element to check for whether it can be recursively documented or not</param>
		/// <returns>True if <paramref name="element"/> can be documented recursively, false otherwise</returns>
		bool IDocumentor.GetElementCanBeDocumentedRecursively(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
				case vsCMElement.vsCMElementInterface:
				case vsCMElement.vsCMElementStruct:
				case vsCMElement.vsCMElementEnum:
				case vsCMElement.vsCMElementModule:
				case vsCMElement.vsCMElementNamespace:
					return true;
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether a project item shall be processed by the documentor according to the user's settings
		/// </summary>
		/// <param name="projectItem">The project item to check for whether it will be processed or not</param>
		/// <returns>True if the project item shall be processed according to the user's settings, false otherwise</returns>
		bool IDocumentor.GetShallBeProcessed(ProjectItem projectItem) {
			if (Array.BinarySearch(extensions, Path.GetExtension(projectItem.Name).ToLower(CultureInfo.InvariantCulture)) < 0) {
				return false;
			}
			return !projectItem.Name.EndsWith(
				global::Simplabs.NetDocumentor.Properties.Resources.BasicDesignerGeneratedFilesNamePostfix,
				StringComparison.InvariantCultureIgnoreCase
			) || !global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicIgnoreDesignerGeneratedFiles;
		}

		/// <summary>
		/// Documents a class
		/// </summary>
		/// <param name="theElement">The class to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the class shall be shown in the code editor after insertion</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theClass"/>. This is used if existing comments are set to be retained and <paramref name="theClass"/> itself has a comment. In this case, the mebers of <paramref name="theClass"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentClass(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			CodeClass2 theClass = (CodeClass2)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReferenceBaseTypes) {
					DocumentAsTypeReferences(theClass.Bases, comment);
					DocumentAsTypeReferences(theClass.ImplementedInterfaces, comment);
				}
				if (theClass.IsGeneric) {
					DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theClass.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a function
		/// </summary>
		/// <param name="theElement">The function to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the function shall be shown in the code editor after insertion</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentFunction(CodeElement2 theElement) {
			CodeFunction2 theFunction = (CodeFunction2)theElement;
			DocumentationComment comment;
			if (theFunction.FunctionKind != vsCMFunction.vsCMFunctionConstructor) {
				comment = GetDocumentationComment(
					theElement,
					global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReuseBaseTypeDocumentation
				);
			} else {
				comment = GetDocumentationComment(theElement);
			}
			if (!comment.Summary.HasValue && theFunction.FunctionKind == vsCMFunction.vsCMFunctionConstructor && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicFunctionsAddSummaryIfConstructor) {
				if (theFunction.IsShared) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Format(
						CultureInfo.CurrentUICulture,
						global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.StaticConstructorSummaryText,
						((CodeElement2)theFunction.Parent).FullName
					));
				} else {
					comment.Summary = new DocumentationComment.SummaryElement(string.Format(
						CultureInfo.CurrentUICulture,
						global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ConstructorSummaryText,
						((CodeElement2)theFunction.Parent).FullName
					));
				}
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			DocumentAsParameters(theFunction.Parameters, comment);
			if (theFunction.IsGeneric) {
				DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
			}
			if (!comment.Returns.HasValue && theFunction.Type.AsFullName != "System.Void") {
				comment.Returns = new DocumentationComment.ReturnsElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents a property
		/// </summary>
		/// <param name="theElement">The property to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the property shall be shown in the code editor after insertion</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentProperty(CodeElement2 theElement) {
			CodeProperty theProperty = (CodeProperty)theElement;
			DocumentationComment comment = GetDocumentationComment(
				theElement,
				global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReuseBaseTypeDocumentation
			);
			if (!comment.Summary.HasValue && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicPropertiesAddSummary) {
				if (theProperty.Getter != null && theProperty.Setter != null) {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ReadWritePropertySummaryText);
				} else if (theProperty.Getter != null) {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ReadOnlyPropertySummaryText);
				} else {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.WriteOnlyPropertySummaryText);
				}
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			if (theProperty.Getter != null) {
				List<CodeElement> parameters = new List<CodeElement>();
				foreach (CodeElement element in theProperty.Getter.Collection) {
					if (element.Kind == vsCMElement.vsCMElementParameter) {
						parameters.Add(element);
					}
				}
				DocumentAsParameters(parameters, comment);
			} else if (theProperty.Setter != null) {
				List<CodeElement> parameters = new List<CodeElement>();
				foreach (CodeElement element in theProperty.Setter.Collection) {
					if (element.Kind == vsCMElement.vsCMElementParameter) {
						parameters.Add(element);
					}
				}
				DocumentAsParameters(parameters, comment);
			}
			if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicPropertiesAddValueElement) {
				comment.Value = new DocumentationComment.ValueElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents a variable
		/// </summary>
		/// <param name="theElement">The variable to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the variable shall be shown in the code editor after insertion</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentField(CodeElement2 theElement) {
			DocumentationComment comment = GetDocumentationComment(theElement);
			if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents an event
		/// </summary>
		/// <param name="theElement">The event to document</param>
		private DocumentationComment DocumentEvent(CodeElement2 theElement) {
			DocumentationComment comment = GetDocumentationComment(
				theElement,
				global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReuseBaseTypeDocumentation
			);
			if (!comment.Summary.HasValue && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicEventsAddSummary) {
				comment.Summary = new DocumentationComment.SummaryElement(
					global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.EventsSummaryText
				);
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents an interface
		/// </summary>
		/// <param name="theElement">The interface to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the interface shall be shown in the code editor after insertion</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theInterface"/>. This is used if existing comments are set to be retained and <paramref name="theInterface"/> itself has a comment. In this case, the mebers of <paramref name="theInterface"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentInterface(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			CodeInterface2 theInterface = (CodeInterface2)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReferenceBaseTypes) {
					DocumentAsTypeReferences(theInterface.Bases, comment);
				}
				if (theInterface.IsGeneric) {
					DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theInterface.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a struct
		/// </summary>
		/// <param name="theElement">The struct to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the struct shall be shown in the code editor after insertion</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theStruct"/>. This is used if existing comments are set to be retained and <paramref name="theStruct"/> itself has a comment. In this case, the mebers of <paramref name="theStruct"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentStruct(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			CodeStruct2 theStruct = (CodeStruct2)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicReferenceBaseTypes) {
					DocumentAsTypeReferences(theStruct.Bases, comment);
				}
				if (theStruct.IsGeneric) {
					DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theStruct.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents an enum
		/// </summary>
		/// <param name="theElement">The enum to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the enum shall be shown in the code editor after insertion</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theEnum"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theEnum"/>. This is used if existing comments are set to be retained and <paramref name="theEnum"/> itself has a comment. In this case, the mebers of <paramref name="theEnum"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentEnum(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			CodeEnum theEnum = (CodeEnum)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theEnum.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a delegate
		/// </summary>
		/// <param name="theElement">The delegate to document</param>
		/// <param name="moveToDeclaration">Specifies whether the declaration position of the enum shall be shown in the code editor after insertion</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentDelegate(CodeElement2 theElement) {
			CodeDelegate2 theDelegate = (CodeDelegate2)theElement;
			DocumentationComment comment = GetDocumentationComment(theElement);
			if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			DocumentAsParameters(theDelegate.Parameters, comment);
			if (!comment.Returns.HasValue && !string.IsNullOrEmpty(theDelegate.Type.AsFullName)) {
				comment.Returns = new DocumentationComment.ReturnsElement(string.Empty);
			}
			if (theDelegate.IsGeneric) {
				DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}


		/// <summary>
		/// Documents a module
		/// </summary>
		/// <param name="theElement">The module to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentModule(CodeElement2 theElement) {
			DocumentationComment comment = GetDocumentationComment(theElement);
			comment.LinePrefix = "''' ";
			if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			foreach (CodeElement2 element in theElement.Children) {
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
					DocumentElement(element, true, true);
				} else {
					DocumentElement(element, true, false);
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a Namespace
		/// </summary>
		/// <param name="theElement">The namespace to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theEnum"/> and all of its members recursively or not</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentNamespace(CodeElement2 theElement, bool documentRecursively) {
			CodeNamespace theNamespace = (CodeNamespace)theElement;
			if (documentRecursively) {
				foreach (CodeElement2 element in theNamespace.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.BasicRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return null;
		}

	}

}