using System;
using System.Collections.Generic;
using System.Text;
using EnvDTE;
using EnvDTE80;
using System.Windows.Forms;
using System.Globalization;
using Microsoft.VisualStudio.Shell.Interop;
using Simplabs.NetDocumentor.IdeIntegration;
using System.ComponentModel;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// Delegate for the <see cref="Simplabs.NetDocumentor.Documentation.CompletionRunner.ItemDocumented"/> event
	/// </summary>
	/// <param name="projectItem">The project item that has been documented</param>
	internal delegate void ProjectItemDocumentedHandler(ProjectItem projectItem);

	/// <summary>
	/// Worker that performs the completions of all documentation in a project
	/// </summary>
	internal class CompletionRunner {

		/// <summary>
		/// The project to complete the documentation for
		/// </summary>
		private Project theProject;

		/// <summary>
		/// The documentor that is used to document the project
		/// </summary>
		private IDocumentor theDocumentor;

		/// <summary>
		/// The code elements that already have been documented during this run
		/// </summary>
		private List<CodeElement> documentedElements;

		/// <summary>
		/// The thread the completion runner runs in
		/// </summary>
		private BackgroundWorker thread;

		/// <summary>
		/// The number of project items that will be documented
		/// </summary>
		private int? numberOfItemsToDocument;

		/// <summary>
		/// Initializes a new CompletionRunner
		/// </summary>
		internal CompletionRunner(Project project) {
			theProject = project;
			theDocumentor = DocumentorFactory.GetDocumentor(theProject.CodeModel.Language);
			theDocumentor.SuppressUI = true;
			documentedElements = new List<CodeElement>();
		}

		/// <summary>
		/// Event that is fired whenever a project item is documented
		/// </summary>
		internal event ProjectItemDocumentedHandler ItemDocumented;

		/// <summary>
		/// Event that is fired whenever a project item is documented
		/// </summary>
		internal event CodeElementDocumentedHandler ElementDocumented;

		/// <summary>
		/// Gets the number of project items that will be documented
		/// </summary>
		internal int NumberOfItemsToDocument {
			get {
				if (!numberOfItemsToDocument.HasValue) {
					numberOfItemsToDocument = 0;
					CountProjectItemsToDocument(theProject.ProjectItems, ref numberOfItemsToDocument);
				}
				return numberOfItemsToDocument.Value;
			}
		}

		/// <summary>
		/// Counts the project items that will be processed by the runner
		/// </summary>
		/// <param name="items">The direct items of the project (these mark the starting point for further, recursive calls)</param>
		/// <param name="counter">The result</param>
		private void CountProjectItemsToDocument(ProjectItems items, ref int? counter) {
			foreach (ProjectItem projectItem in items) {
				CountProjectItemsToDocument(projectItem.ProjectItems, ref counter);
				if (theDocumentor.GetShallBeProcessed(projectItem)) {
					counter++;
				}
			}
		}

		/// <summary>
		/// Documents all code elements in the project items <paramref name="projectItems"/> and their subitems
		/// </summary>
		/// <param name="projectItems">The project items to document the code elements for</param>
		private void DocumentProjectItems(ProjectItems projectItems, int counter) {
			foreach (ProjectItem projectItem in projectItems) {
				if (theDocumentor.GetShallBeProcessed(projectItem)) {
					try {
						FileCodeModel2 codeModel = projectItem.FileCodeModel as FileCodeModel2;
						DocumentElements(codeModel.CodeElements);
						if (projectItem.IsDirty) {
							projectItem.Save(null);
						}
						counter++;
						thread.ReportProgress((int)Math.Floor((decimal)numberOfItemsToDocument / counter));
						if (ItemDocumented != null) {
							ItemDocumented(projectItem);
						}
					} catch (Exception e) {
						Console.Write(e.ToString());
						//catch all exceptions to continue on errors
					}
				}
				if (thread.CancellationPending) {
					break;
				}
				DocumentProjectItems(projectItem.ProjectItems, counter);
			}
		}

		/// <summary>
		/// Documents the code elements <paramref name="elements"/> and its children
		/// </summary>
		/// <param name="elements">The code elements to document</param>
		private void DocumentElements(CodeElements elements) {
			foreach (CodeElement2 element in elements) {
				if (!documentedElements.Contains(element)) {
					theDocumentor.DocumentElement(element, true);
				}
				if (element.Kind != vsCMElement.vsCMElementProperty) {
					DocumentElements(element.Children);
				}
				if (thread.CancellationPending) {
					break;
				}
			}
		}

		/// <summary>
		/// Starts the process
		/// </summary>
		/// <param name="insertTaskShortcuts">Specifies whether task shortcuts are to be added to documentation that requires manual input by the user</param>
		/// <param name="workerThread">The thread the completion runner runs in</param>
		internal void Run(bool insertTaskShortcuts, BackgroundWorker workerThread) {
			thread = workerThread;
			thread.WorkerReportsProgress = true;
			Window taskWindow = null;
			DocumentationTaskProvider taskProvider = null;
			if (insertTaskShortcuts) {
				taskWindow = theProject.DTE.Windows.Item(EnvDTE.Constants.vsWindowKindTaskList);
				taskProvider = new DocumentationTaskProvider();
			}
			theDocumentor.ElementDocumented += new CodeElementDocumentedHandler(delegate(CodeElement codeElement, bool isDocumentationIncomplete) {
				if (insertTaskShortcuts) {
					taskProvider.AddNewTask(codeElement);
					documentedElements.Add(codeElement);
				}
				if (ElementDocumented != null) {
					ElementDocumented(codeElement, isDocumentationIncomplete);
				}
			});
			DocumentProjectItems(theProject.ProjectItems, 0);
			if (insertTaskShortcuts) {
				taskProvider.ShowAllTasks();
				taskWindow = theProject.DTE.Windows.Item(EnvDTE.Constants.vsWindowKindTaskList);
				taskWindow.Activate();
				taskWindow.Visible = true;
			}
		}

	}

}