using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.VCCodeModel;
using EnvDTE;
using System.Text.RegularExpressions;
using System.Drawing;
using System.Diagnostics;
using System.Globalization;
using Simplabs.NetDocumentor.Settings;
using System.IO;
using Simplabs.NetDocumentor.Documentation;
using EnvDTE80;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// Documentor for C, C++ and C++/CLI code
	/// </summary>
	internal class CppDocumentor : IDocumentor {

		/// <summary>
		/// The language prefix
		/// </summary>
		/// <remarks>This prefix is used to retrieve confiiguration settings etc. for this language</remarks>
		internal const string LanguagePrefix = "Cpp";

		/// <summary>
		/// The file extensions the documentor supports
		/// </summary>
		private string[] extensions = { ".h" };

		/// <summary>
		/// Specifies whether the documentor suppresses the UI
		/// </summary>
		private bool suppressUI;

		/// <summary>
		/// Initializes a new Simplabs.NetDocumentor
		/// </summary>
		internal CppDocumentor() {
		}

		/// <summary>
		/// Specifies whether the documentor should suppress the UI
		/// </summary>
		bool IDocumentor.SuppressUI {
			get {
				return suppressUI;
			}
			set {
				suppressUI = value;
			}
		}

		/// <summary>
		/// Event that is fired whenever a code element is documented
		/// </summary>
		public event CodeElementDocumentedHandler ElementDocumented;

		/// <summary>
		/// Gets a documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.
		/// </summary>
		/// <param name="theElement">The code element to get the documentation comment for</param>
		/// <returns>A documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.</returns>
		private DocumentationComment GetDocumentationComment(CodeElement2 theElement) {
			return GetDocumentationComment(theElement, false);
		}

		/// <summary>
		/// Gets a documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.
		/// </summary>
		/// <param name="theElement">The code element to get the documentation comment for</param>
		/// <param name="resuseBaseTypeDocumentation">Specifies whether base type documentations are to be reused for override, hides or interface implementations</param>
		/// <returns>A documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.</returns>
		/// <remarks>If the code element is an overridden method, a method giding another from the base type or an element implementing an interface and does not have a documentation comment itself, the documentation comment of the matching base type's member will be returned if available</remarks>
		private DocumentationComment GetDocumentationComment(CodeElement2 theElement, bool resuseBaseTypeDocumentation) {
			VCCodeElement element = (VCCodeElement)theElement;
			if (SourceCodeServices.HasDocumentationComment(element, vsCMWhere.vsCMWhereDeclaration) || SourceCodeServices.HasDocumentationComment(element, vsCMWhere.vsCMWhereDefinition)) {
				return SourceCodeServices.ParseDocumentationComment(element, "/// ");
			}
			if (resuseBaseTypeDocumentation && (
				element.Kind == vsCMElement.vsCMElementFunction || element.Kind == vsCMElement.vsCMElementEvent || element.Kind == vsCMElement.vsCMElementProperty
			)) {
				VCCodeElement matchingBaseTypeElement = null;
				CodeType containingType = null;
				string elementPrototype = string.Empty;
				vsCMElement searchedElementKind = vsCMElement.vsCMElementOther;
				if (element.Kind == vsCMElement.vsCMElementFunction) {
					containingType = (CodeType)(((VCCodeFunction)element).Parent);
					searchedElementKind = vsCMElement.vsCMElementFunction;
					elementPrototype = ((VCCodeFunction)element).get_Prototype(
						(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
					);
				} else if (element.Kind == vsCMElement.vsCMElementEvent) {
					containingType = (CodeType)(((VCCodeEvent)element).Parent);
					searchedElementKind = vsCMElement.vsCMElementEvent;
					elementPrototype = ((VCCodeEvent)element).TypeString + ":" + ((VCCodeEvent)element).Name;
				} else if (element.Kind == vsCMElement.vsCMElementProperty) {
					containingType = (CodeType)(((VCCodeProperty)element).Parent);
					searchedElementKind = vsCMElement.vsCMElementProperty;
					elementPrototype = ((VCCodeProperty)element).get_Prototype(
						(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
					);
				}
				if (element.Name.Contains(".")) {
					elementPrototype = Regex.Replace(elementPrototype, @"\s[^\.]+\.", " ", RegexOptions.IgnoreCase);
				}
				foreach (VCCodeBase baseType in containingType.Bases) {
					if (baseType.Class != null) {
						matchingBaseTypeElement = FindMatchingBaseTypeMember((VCCodeElement)element, baseType.Class, searchedElementKind, elementPrototype);
						if (matchingBaseTypeElement != null) {
							return SourceCodeServices.ParseDocumentationComment(matchingBaseTypeElement, "/// ");
						}
					}
				}
				if (containingType.Kind == vsCMElement.vsCMElementClass) {
					foreach (VCCodeBase implementedInterface in ((VCCodeClass)containingType).ImplementedInterfaces) {
						if (implementedInterface.Class != null) {
							matchingBaseTypeElement = FindMatchingBaseTypeMember(element, implementedInterface.Class, searchedElementKind, elementPrototype);
							if (matchingBaseTypeElement != null) {
								return SourceCodeServices.ParseDocumentationComment(matchingBaseTypeElement, "/// ");
							}
						}
					}
				}
				if (containingType.Kind == vsCMElement.vsCMElementStruct) {
					foreach (VCCodeBase implementedInterface in ((VCCodeStruct)containingType).ImplementedInterfaces) {
						if (implementedInterface.Class != null) {
							matchingBaseTypeElement = FindMatchingBaseTypeMember(element, implementedInterface.Class, searchedElementKind, elementPrototype);
							if (matchingBaseTypeElement != null) {
								return SourceCodeServices.ParseDocumentationComment(matchingBaseTypeElement, "/// ");
							}
						}
					}
				}
			}
			return new DocumentationComment("/// ");
		}

		/// <summary>
		/// Finds a code element matching the signature of <paramref name="element"/> in type <paramref name="theType"/>
		/// </summary>
		/// <param name="element">The code element to find a matching code element for</param>
		/// <param name="theType">The type to find the matching code element in (all base types of the type are also included in the search)</param>
		/// <param name="searchedElementKind">The kind of element that is searched for</param>
		/// <param name="elementPrototype">The prototype of the element that is searched for</param>
		/// <returns>The code element matching the signature of <paramref name="element"/> or null if no such element exists</returns>
		private VCCodeElement FindMatchingBaseTypeMember(VCCodeElement element, CodeType theType, vsCMElement searchedElementKind, string elementPrototype) {
			if (element.Kind != vsCMElement.vsCMElementFunction && element.Kind != vsCMElement.vsCMElementEvent && element.Kind != vsCMElement.vsCMElementProperty) {
				return null;
			}
			foreach (VCCodeElement baseTypeElement in theType.Members) {
				if (baseTypeElement.Kind == searchedElementKind) {
					string baseElementPrototype = string.Empty;
					if (baseTypeElement.Kind == vsCMElement.vsCMElementFunction) {
						baseElementPrototype = ((VCCodeFunction)baseTypeElement).get_Prototype((int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes));
					} else if (baseTypeElement.Kind == vsCMElement.vsCMElementEvent) {
						baseElementPrototype = ((VCCodeEvent)baseTypeElement).TypeString + ":" + ((VCCodeEvent)baseTypeElement).Name;
					} else if (baseTypeElement.Kind == vsCMElement.vsCMElementProperty) {
						baseElementPrototype = ((VCCodeProperty)baseTypeElement).get_Prototype((int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes));
					}
					if (!string.IsNullOrEmpty(baseElementPrototype) && baseElementPrototype.Equals(elementPrototype)) {
						return (VCCodeElement)baseTypeElement;
					}
				}
			}
			foreach (VCCodeBase superBaseType in theType.Bases) {
				if (superBaseType.Class != null) {
					VCCodeElement upperLevelmatch = FindMatchingBaseTypeMember(element, superBaseType.Class, searchedElementKind, elementPrototype);
					if (upperLevelmatch != null) {
						return upperLevelmatch;
					}
				}
			}
			if (theType.Kind == vsCMElement.vsCMElementClass) {
				foreach (VCCodeBase theInterface in ((VCCodeClass)theType).ImplementedInterfaces) {
					if (theInterface.Class != null) {
						VCCodeElement upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface.Class, searchedElementKind, elementPrototype);
						if (upperLevelmatch != null) {
							return upperLevelmatch;
						}
					}
				}
			} else if (theType.Kind == vsCMElement.vsCMElementStruct) {
				foreach (VCCodeBase theInterface in ((VCCodeStruct)theType).ImplementedInterfaces) {
					if (theInterface.Class != null) {
						VCCodeElement upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface.Class, searchedElementKind, elementPrototype);
						if (upperLevelmatch != null) {
							return upperLevelmatch;
						}
					}
				}
			} else if (theType.Kind == vsCMElement.vsCMElementInterface) {
				foreach (VCCodeBase theInterface in ((VCCodeInterface)theType).Bases) {
					if (theInterface.Class != null) {
						VCCodeElement upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface.Class, searchedElementKind, elementPrototype);
						if (upperLevelmatch != null) {
							return upperLevelmatch;
						}
					}
				}
			}
			return null;
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as references
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		private void DocumentAsReferences(CodeElements elements, DocumentationComment comment) {
			foreach (CodeElement2 element in elements) {
				if (!element.FullName.Equals(typeof(object).FullName.Replace(".", "::"), StringComparison.InvariantCultureIgnoreCase)) {
					DocumentationComment.ReferenceElement referenceElement = new DocumentationComment.ReferenceElement(element.FullName);
					comment.References.Add(referenceElement);
				}
			}
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as type parameters
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		private void DocumentAsTypeParameters(List<string> elements, DocumentationComment comment) {
			foreach (string typeParameter in elements) {
				DocumentationComment.TypeParameterElement typeParameterElement = new DocumentationComment.TypeParameterElement(typeParameter, string.Empty);
				comment.TypeParameters.Add(typeParameterElement);
			}
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as parameters
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		private void DocumentAsParameters(CodeElements elements, DocumentationComment comment) {
			foreach (CodeElement2 parameter in elements) {
				DocumentationComment.ParameterElement parameterElement = new DocumentationComment.ParameterElement(parameter.Name, string.Empty);
				comment.Parameters.Add(parameterElement);
			}
		}

		/// <summary>
		/// Applies the documentation comment <paramref name="comment"/> to code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The code element to apply the documentation comment to</param>
		/// <param name="comment">The documentation comment to apply to the code element <paramref name="element"/></param>
		private void ApplyDocumentation(CodeElement2 theElement, DocumentationComment comment) {
			VCCodeElement element = (VCCodeElement)theElement;
			SourceCodeServices.InsertDocumentation(element, comment.ToString());
			if (!suppressUI) {
				vsCMWhere elementPortion = vsCMWhere.vsCMWhereDefault;
				if (global::Simplabs.NetDocumentor.Properties.LanguageSpecificSettings.Default.CppDocumentAtDeclaration) {
					elementPortion = vsCMWhere.vsCMWhereDeclaration;
				} else if (global::Simplabs.NetDocumentor.Properties.LanguageSpecificSettings.Default.CppDocumentAtDefinition) {
					elementPortion = vsCMWhere.vsCMWhereDefinition;
				}
				SourceCodeServices.ShowDocumentation(element, elementPortion);
			}
		}

		/// <summary>
		/// Initializes the documentor
		/// </summary>
		void IDocumentor.Initialize() {
			suppressUI = false;
		}

		/// <summary>
		/// Generates the XML documentation structure for a given code <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to document</param>
		/// <param name="documentRecursively">Set to true to document <paramref name="element"/> and alll of its members recursively or set to false to document just <paramref name="element"/></param>
		/// <exception cref="Simplabs.NetDocumentor.Documentation.DocumentationException">Thrown when an error occures during application of the documentation</exception>
		void IDocumentor.DocumentElement(CodeElement2 element, bool documentRecursively) {
			try {
				DocumentElement(element, documentRecursively, false);
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			}
		}

		/// <summary>
		/// Generates the XML documentation structure for a given code <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to document</param>
		/// <param name="documentRecursively">Set to true to document <paramref name="element"/> and alll of its members recursively or set to false to document just <paramref name="element"/></param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="elemment"/>. This is used if existing comments are set to be retained and <paramref name="element"/> itself has a comment. In this case, the mebers of <paramref name="element"/> still could have no comments.</param>
		private void DocumentElement(CodeElement2 element, bool documentRecursively, bool membersOnly) {
			try {
				bool preserveComment = true;
				if (documentRecursively && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CSharpRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
					preserveComment = false;
				}
				DocumentationComment comment = null;
				switch (element.Kind) {
					case vsCMElement.vsCMElementClass:
						if (preserveComment) {
							comment = DocumentClass(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentClass(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementFunction:
						if (!membersOnly) {
							comment = DocumentFunction(element);
						}
						break;
					case vsCMElement.vsCMElementProperty:
						if (!membersOnly) {
							comment = DocumentProperty(element);
						}
						break;
					case vsCMElement.vsCMElementVariable:
						if (!membersOnly) {
							comment = DocumentField(element);
						}
						break;
					case vsCMElement.vsCMElementEvent:
						if (!membersOnly) {
							comment = DocumentEvent(element);
						}
						break;
					case vsCMElement.vsCMElementInterface:
						if (preserveComment) {
							comment = DocumentInterface(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentInterface(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementStruct:
						if (preserveComment) {
							comment = DocumentStruct(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentStruct(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementEnum:
						if (preserveComment) {
							comment = DocumentEnum(element, documentRecursively, membersOnly);
						} else if (documentRecursively) {
							comment = DocumentEnum(element, documentRecursively, true);
						}
						break;
					case vsCMElement.vsCMElementDelegate:
						if (!membersOnly) {
							comment = DocumentDelegate(element);
						}
						break;
					case vsCMElement.vsCMElementNamespace:
						if (documentRecursively) {
							comment = DocumentNamespace(element, true);
						}
						break;
					default:
						return;
				}
				if (comment != null && ElementDocumented != null) {
					ElementDocumented(element, comment != null ? comment.IsIncomplete : false);
				}
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> will be documented according to the current settings
		/// </summary>
		/// <param name="element">The element to check for whether it will be documented or not</param>
		bool IDocumentor.GetElementWillBeDocumented(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
				case vsCMElement.vsCMElementFunction:
				case vsCMElement.vsCMElementProperty:
				case vsCMElement.vsCMElementVariable:
				case vsCMElement.vsCMElementEvent:
				case vsCMElement.vsCMElementInterface:
				case vsCMElement.vsCMElementStruct:
				case vsCMElement.vsCMElementEnum:
					return true;
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> can be recursively documented
		/// </summary>
		/// <param name="element">The element to check for whether it can be recursively documented or not</param>
		/// <returns>True if <paramref name="element"/> can be documented recursively, false otherwise</returns>
		bool IDocumentor.GetElementCanBeDocumentedRecursively(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
				case vsCMElement.vsCMElementInterface:
				case vsCMElement.vsCMElementStruct:
				case vsCMElement.vsCMElementEnum:
				case vsCMElement.vsCMElementNamespace:
					return true;
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether a project item shall be processed by the documentor according to the user's settings
		/// </summary>
		/// <param name="projectItem">The project item to check for whether it will be processed or not</param>
		/// <returns>True if the project item shall be processed according to the user's settings, false otherwise</returns>
		bool IDocumentor.GetShallBeProcessed(ProjectItem projectItem) {
			if (Array.BinarySearch(extensions, Path.GetExtension(projectItem.Name).ToLower(CultureInfo.InvariantCulture)) < 0) {
				return false;
			}
			return !projectItem.Name.EndsWith(
				global::Simplabs.NetDocumentor.Properties.Resources.CppDesignerGeneratedFilesNamePostfix,
				StringComparison.InvariantCultureIgnoreCase
			) || !global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppIgnoreDesignerGeneratedFiles;
		}

		/// <summary>
		/// Documents a class
		/// </summary>
		/// <param name="theElement">The class to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theClass"/>. This is used if existing comments are set to be retained and <paramref name="theClass"/> itself has a comment. In this case, the mebers of <paramref name="theClass"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentClass(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			VCCodeClass theClass = (VCCodeClass)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReferenceBaseTypes) {
					DocumentAsReferences(theClass.Bases, comment);
					DocumentAsReferences(theClass.ImplementedInterfaces, comment);
				}
				if (theClass.IsTemplate) {
					DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters(theElement), comment);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theClass.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a function
		/// </summary>
		/// <param name="theElement">The function to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentFunction(CodeElement2 theElement) {
			VCCodeFunction theFunction = (VCCodeFunction)theElement;
			DocumentationComment comment;
			if (theFunction.FunctionKind != vsCMFunction.vsCMFunctionConstructor) {
				comment = GetDocumentationComment(
					theElement,
					global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReuseBaseTypeDocumentation
				);
			} else {
				comment = GetDocumentationComment(theElement);
			}
			if (!comment.Summary.HasValue && theFunction.FunctionKind  == vsCMFunction.vsCMFunctionConstructor && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppFunctionsAddSummaryIfConstructor) {
				if (theFunction.IsShared) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Format(
						CultureInfo.CurrentUICulture,
						global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.StaticConstructorSummaryText,
						((VCCodeElement)theFunction.Parent).FullName
					));
				} else {
					comment.Summary = new DocumentationComment.SummaryElement(string.Format(
						CultureInfo.CurrentUICulture,
						global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ConstructorSummaryText,
						((VCCodeElement)theFunction.Parent).FullName
					));
				}
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			DocumentAsParameters(theFunction.Parameters, comment);
			if (theFunction.IsTemplate) {
				DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters((CodeElement2)theElement), comment);
			}
			if (!comment.Returns.HasValue && theFunction.Type.AsFullName != "void") {
				comment.Returns = new DocumentationComment.ReturnsElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents a property
		/// </summary>
		/// <param name="theElement">The property to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentProperty(CodeElement2 theElement) {
			VCCodeProperty theProperty = (VCCodeProperty)theElement;
			DocumentationComment comment = GetDocumentationComment(
				theElement,
				global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReuseBaseTypeDocumentation
			);
			if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppPropertiesAddSummary) {
				if (theProperty.Getter != null && theProperty.Setter != null) {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ReadWritePropertySummaryText);
				} else if (theProperty.Getter != null) {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.ReadOnlyPropertySummaryText);
				} else {
					comment.Summary = new DocumentationComment.SummaryElement(global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.WriteOnlyPropertySummaryText);
				}
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			if (theProperty.Getter != null) {
				DocumentAsParameters(theProperty.Getter.Parameters, comment);
			} else if (theProperty.Setter != null) {
				DocumentAsParameters(theProperty.Setter.Parameters, comment);
			}
			if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppPropertiesAddValueElement) {
				comment.Value = new DocumentationComment.ValueElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents a variable
		/// </summary>
		/// <param name="theElement">The variable to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentField(CodeElement2 theElement) {
			DocumentationComment comment = GetDocumentationComment(theElement);
			if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents an event
		/// </summary>
		/// <param name="theElement">The event to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentEvent(CodeElement2 theElement) {
			DocumentationComment comment = GetDocumentationComment(
				theElement,
				global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReuseBaseTypeDocumentation
			);
			if (!comment.Summary.HasValue && global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppEventsAddSummary) {
				comment.Summary = new DocumentationComment.SummaryElement(
					global::Simplabs.NetDocumentor.Properties.TextSnippetSettings.Default.EventsSummaryText
				);
			} else if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents an interface
		/// </summary>
		/// <param name="theElement">The interface to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theInterface"/>. This is used if existing comments are set to be retained and <paramref name="theInterface"/> itself has a comment. In this case, the mebers of <paramref name="theInterface"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentInterface(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			VCCodeInterface theInterface = (VCCodeInterface)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReferenceBaseTypes) {
					DocumentAsReferences(theInterface.Bases, comment);
				}
				DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters((CodeElement2)theInterface), comment);
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theInterface.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents a struct
		/// </summary>
		/// <param name="theElement">The struct to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theStruct"/>. This is used if existing comments are set to be retained and <paramref name="theStruct"/> itself has a comment. In this case, the mebers of <paramref name="theStruct"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentStruct(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			VCCodeStruct theStruct = (VCCodeStruct)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppReferenceBaseTypes) {
					DocumentAsReferences(theStruct.Bases, comment);
					DocumentAsReferences(theStruct.ImplementedInterfaces, comment);
				}
				DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters((CodeElement2)theStruct), comment);
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theStruct.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment;
		}

		/// <summary>
		/// Documents an enum
		/// </summary>
		/// <param name="theElement">The enum to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theClass"/> and all of its members recursively or not</param>
		/// <param name="membersOnly">Specifies whether to document only the members of <paramref name="theEnum"/>. This is used if existing comments are set to be retained and <paramref name="theEnum"/> itself has a comment. In this case, the mebers of <paramref name="theEnum"/> still could have no comments.</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentEnum(CodeElement2 theElement, bool documentRecursively, bool membersOnly) {
			VCCodeEnum theEnum = (VCCodeEnum)theElement;
			DocumentationComment comment = null;
			if (!membersOnly) {
				comment = GetDocumentationComment(theElement);
				if (!comment.Summary.HasValue) {
					comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
				}
				ApplyDocumentation(theElement, comment);
			}
			if (documentRecursively) {
				foreach (CodeElement2 element in theEnum.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return comment	;
		}

		/// <summary>
		/// Documents a delegate
		/// </summary>
		/// <param name="theElement">The delegate to document</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentDelegate(CodeElement2 theElement) {
			VCCodeDelegate theDelegate = (VCCodeDelegate)theElement;
			DocumentationComment comment = GetDocumentationComment(theElement);
			if (!comment.Summary.HasValue) {
				comment.Summary = new DocumentationComment.SummaryElement(string.Empty);
			}
			DocumentAsParameters(theDelegate.Parameters, comment);
			if (theDelegate.Type.AsFullName != "void") {
				comment.Returns = new DocumentationComment.ReturnsElement(string.Empty);
			}
			DocumentAsTypeParameters(SourceCodeServices.GetTemplateParameters((CodeElement2)theDelegate), comment);
			ApplyDocumentation(theElement, comment);
			return comment;
		}

		/// <summary>
		/// Documents a Namespace
		/// </summary>
		/// <param name="theElement">The namespace to document</param>
		/// <param name="documentRecursively">specifies whether to document <paramref name="theEnum"/> and all of its members recursively or not</param>
		/// <returns>The documentation comment</returns>
		private DocumentationComment DocumentNamespace(CodeElement2 theElement, bool documentRecursively) {
			VCCodeNamespace theNamespace = (VCCodeNamespace)theElement;
			if (documentRecursively) {
				foreach (CodeElement2 element in theNamespace.Members) {
					if (global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.CppRetainExistingDocumentation && SourceCodeServices.HasComment(element)) {
						DocumentElement(element, true, true);
					} else {
						DocumentElement(element, true, false);
					}
				}
			}
			return null;
		}

	}

}