using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using System.Collections;
using System.Xml;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// Structure that wraps documentation comments
	/// </summary>
	internal class DocumentationComment {

		/// <summary>
		/// The line prefix to use (e.g. <code>///</code>)
		/// </summary>
		private string linePrefix;

		/// <summary>
		/// The summary element
		/// </summary>
		private SummaryElement? summary;

		/// <summary>
		/// The returns element
		/// </summary>
		private ReturnsElement? returns;

		/// <summary>
		/// The value element
		/// </summary>
		private ValueElement? propertyValue;

		/// <summary>
		/// The parameter elements
		/// </summary>
		private ParameterElementList parameters = new ParameterElementList();

		/// <summary>
		/// The type parameter elements
		/// </summary>
		private TypeParameterElementList typeParameters = new TypeParameterElementList();

		/// <summary>
		/// The exception elements
		/// </summary>
		private ExceptionElementList exceptions = new ExceptionElementList();

		/// <summary>
		/// The reference elements
		/// </summary>
		private ReferenceElementList references = new ReferenceElementList();

		/// <summary>
		/// All elements in the documentation comment that are not directly supported by .Net Documentor
		/// </summary>
		/// <remarks>These are only elements that are retrieved from existing comments, rather than elements generated by .Net Documentor</remarks>
		private List<UnsupportedElement> unsupportedElements;

		/// <summary>
		/// Initializes a new DocumentationComment
		/// </summary>
		internal DocumentationComment() {
			unsupportedElements = new List<UnsupportedElement>();
		}

		/// <summary>
		/// Initializes a new DocumentationComment
		/// </summary>
		internal DocumentationComment(string commentLinePrefix) : this() {
			linePrefix = commentLinePrefix;
		}

		/// <summary>
		/// Gets or sets the line prefix to use
		/// </summary>
		internal string LinePrefix {
			get {
				return linePrefix;
			}
			set {
				linePrefix = value;
			}
		}

		/// <summary>
		/// Gets the summary element
		/// </summary>
		internal SummaryElement? Summary {
			get {
				return summary;
			}
			set {
				summary = value;
			}
		}

		/// <summary>
		/// Gets the returns element
		/// </summary>
		internal ReturnsElement? Returns {
			get {
				return returns;
			}
			set {
				returns = value;
			}
		}

		/// <summary>
		/// Gets the value element
		/// </summary>
		internal ValueElement? Value {
			get {
				return propertyValue;
			}
			set {
				propertyValue = value;
			}
		}

		/// <summary>
		/// Gets the parameter elements
		/// </summary>
		internal ParameterElementList Parameters {
			get {
				return parameters;
			}
		}

		/// <summary>
		/// Gets the type parameter elements
		/// </summary>
		internal TypeParameterElementList TypeParameters {
			get {
				return typeParameters;
			}
		}

		/// <summary>
		/// Gets the exception elements
		/// </summary>
		internal ExceptionElementList Exceptions {
			get {
				return exceptions;
			}
		}

		/// <summary>
		/// Gets the reference elements
		/// </summary>
		internal ReferenceElementList References {
			get {
				return references;
			}
		}

		/// <summary>
		/// Gets whether the documentation comment is incomplete and requires manual input
		/// </summary>
		internal bool IsIncomplete {
			get {
				if (summary.HasValue && string.IsNullOrEmpty(summary.Value.Text)) {
					return true;
				}
				if (returns.HasValue && string.IsNullOrEmpty(returns.Value.Text)) {
					return true;
				}
				if (propertyValue.HasValue && string.IsNullOrEmpty(propertyValue.Value.Text)) {
					return true;
				}
				foreach (ParameterElement parameter in parameters) {
					if (string.IsNullOrEmpty(parameter.Description)) {
						return true;
					}
				}
				foreach (TypeParameterElement typeParameter in typeParameters) {
					if (string.IsNullOrEmpty(typeParameter.Description)) {
						return true;
					}
				}
				foreach (ExceptionElement exception in exceptions) {
					if (string.IsNullOrEmpty(exception.Description)) {
						return true;
					}
				}
				return false;
			}
		}

		/// <summary>
		/// Sets the unsupported elements of the documentation comment
		/// </summary>
		/// <param name="nodes">The Xml nodes representing the unsupported elements</param>
		internal void SetUnsupportedElements(XmlNodeList nodes) {
			foreach (XmlNode node in nodes) {
				unsupportedElements.Add(new UnsupportedElement(node));
			}
		}

		/// <summary>
		/// Gets the string representation of the documentation comment
		/// </summary>
		/// <returns>The string representation of the documentation comment</returns>
		public override string ToString() {
			string comment = string.Empty;
			if (summary.HasValue) {
				comment += summary.Value.ToString(linePrefix);
			}
			comment += parameters.ToString(linePrefix);
			comment += typeParameters.ToString(linePrefix);
			if (returns.HasValue) {
				comment += returns.Value.ToString(linePrefix);
			}
			if (propertyValue.HasValue) {
				comment += propertyValue.Value.ToString(linePrefix);
			}
			comment += references.ToString(linePrefix);
			comment += exceptions.ToString(linePrefix);
			foreach (UnsupportedElement unsupportedNode in unsupportedElements) {
				comment += unsupportedNode.ToString(linePrefix);
			}
			return comment;
		}

		/// <summary>
		/// A summmary element
		/// </summary>
		internal struct SummaryElement {

			/// <summary>
			/// The text of the summary element
			/// </summary>
			private string text;

			/// <summary>
			/// Initializes a new SummaryElement
			/// </summary>
			/// <param name="summaryText">The text of the summary element</param>
			internal SummaryElement(string summaryText) {
				text = summaryText.Trim();
			}

			/// <summary>
			/// Gets or sets the text of the summary element
			/// </summary>
			internal string Text {
				get {
					return text;
				}
				set {
					text = value.Trim();
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				SummaryElement summaryElement = (SummaryElement)otherObject;
				return text == summaryElement.Text;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(SummaryElement element1, SummaryElement element2) {
				return element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(SummaryElement element1, SummaryElement element2) {
				return !element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return text.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}<summary>{1}{2}{3}{4}{5}</summary>",
					linePrefix,
					Environment.NewLine,
					linePrefix,
					text.Replace(Environment.NewLine, Environment.NewLine + linePrefix),
					Environment.NewLine,
					linePrefix
				);
			}

		}

		/// <summary>
		/// A returns element
		/// </summary>
		internal struct ReturnsElement {

			/// <summary>
			/// The text of the returns element
			/// </summary>
			private string text;

			/// <summary>
			/// Initializes a new ReturnsElement
			/// </summary>
			/// <param name="returnsText">The text of the returns element</param>
			internal ReturnsElement(string returnsText) {
				text = returnsText.Trim();
			}

			/// <summary>
			/// Gets or sets the text of the returns element
			/// </summary>
			internal string Text {
				get {
					return text;
				}
				set {
					text = value.Trim();
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				ReturnsElement returnsElement = (ReturnsElement)otherObject;
				return text == returnsElement.Text;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(ReturnsElement element1, ReturnsElement element2) {
				return element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(ReturnsElement element1, ReturnsElement element2) {
				return !element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return text.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<returns>{2}</returns>",
					Environment.NewLine,
					linePrefix,
					text.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

		/// <summary>
		/// A value element
		/// </summary>
		internal struct ValueElement {

			/// <summary>
			/// The text of the value element
			/// </summary>
			private string text;

			/// <summary>
			/// Initializes a new ValueElement
			/// </summary>
			/// <param name="returnsText">The text of the value element</param>
			internal ValueElement(string returnsText) {
				text = returnsText.Trim();
			}

			/// <summary>
			/// Gets or sets the text of the value element
			/// </summary>
			internal string Text {
				get {
					return text;
				}
				set {
					text = value.Trim();
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				ValueElement valueElement = (ValueElement)otherObject;
				return text == valueElement.Text;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(ValueElement element1, ValueElement element2) {
				return element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(ValueElement element1, ValueElement element2) {
				return !element1.Text.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return text.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<value>{2}</value>",
					Environment.NewLine,
					linePrefix,
					text.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

		/// <summary>
		/// A parameter element
		/// </summary>
		internal struct ParameterElement {

			/// <summary>
			/// The name of the parameter
			/// </summary>
			private string name;

			/// <summary>
			/// The description of the parameter
			/// </summary>
			private string description;

			/// <summary>
			/// Initializes a new ParameterElement
			/// </summary>
			/// <param name="parameterName">The name of the parameter</param>
			/// <param name="parameterDescription">The description of the parameter</param>
			internal ParameterElement(string parameterName, string parameterDescription) {
				if (!string.IsNullOrEmpty(parameterName)) {
					name = parameterName.Trim();
				} else {
					name = parameterName;
				}
				if (!string.IsNullOrEmpty(parameterDescription)) {
					description = parameterDescription.Trim();
				} else {
					description = parameterDescription;
				}
			}

			/// <summary>
			/// Gets or sets the name of the parameter
			/// </summary>
			internal string ParameterName {
				get {
					return name;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						name = value.Trim();
					} else {
						name = value;
					}
				}
			}

			/// <summary>
			/// Gets or sets the description of the parameter
			/// </summary>
			internal string Description {
				get {
					return description;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						description = value.Trim();
					} else {
						description = value;
					}
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				ParameterElement parameterElement = (ParameterElement)otherObject;
				if (name != parameterElement.ParameterName) {
					return false;
				}
				return description == parameterElement.Description;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(ParameterElement element1, ParameterElement element2) {
				return element1.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(ParameterElement element1, ParameterElement element2) {
				return !element1.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return name.GetHashCode() ^ description.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<param name=\"{2}\">{3}</param>",
					Environment.NewLine,
					linePrefix,
					name,
					description.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

		/// <summary>
		/// A list of <see cref="Simplabs.NetDocumentor.Documentation.ParameterElement"/> instances
		/// </summary>
		internal class ParameterElementList : IEnumerable<ParameterElement> {

			/// <summary>
			/// The actual list that stores the elements
			/// </summary>
			private List<ParameterElement> list;

			/// <summary>
			/// Initializes a new ParameterElementList
			/// </summary>
			internal ParameterElementList() {
				list = new List<ParameterElement>();
			}

			/// <summary>
			/// Gets the parameter at index <paramref name="i"/>
			/// </summary>
			/// <param name="i">The index to get the parameter from</param>
			/// <returns>The parameter at index <paramref name="i"/></returns>
			internal ParameterElement this[int i] {
				get {
					return list[i];
				}
			}

			/// <summary>
			/// Gets the parameter with name <paramref name="parameterName"/>
			/// </summary>
			/// <param name="parameterName">The name of the parameter to get</param>
			/// <returns>The parameter with name <paramref name="parameterName"/> or nulll if no such parameter exists</returns>
			internal ParameterElement? this[string parameterName] {
				get {
					foreach (ParameterElement parameter in list) {
						if (parameter.ParameterName == parameterName) {
							return parameter;
						}
					}
					return null;
				}
			}

			/// <summary>
			/// Adds a parameter to the list
			/// </summary>
			/// <param name="parameter">The parameter to add</param>
			internal void Add(ParameterElement parameter) {
				foreach (ParameterElement element in list) {
					if (element.ParameterName == parameter.ParameterName) {
						return;
					}
				}
				list.Add(parameter);
			}

			/// <summary>
			/// Removes all Parameter from the list
			/// </summary>
			internal void Clear() {
				list.Clear();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			IEnumerator<ParameterElement> IEnumerable<ParameterElement>.GetEnumerator() {
				return list.GetEnumerator();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
				return ((IEnumerable)list).GetEnumerator();
			}

			/// <summary>
			/// Gets the string representation of the parameters in the list
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the parameters in the list</returns>
			internal string ToString(string linePrefix) {
				string comment = string.Empty;
				foreach (ParameterElement parameter in list) {
					comment += parameter.ToString(linePrefix);
				}
				return comment;
			}

		}

		/// <summary>
		/// A type parameter element
		/// </summary>
		internal struct TypeParameterElement {

			/// <summary>
			/// The name of the type parameter
			/// </summary>
			private string name;

			/// <summary>
			/// The description of the type parameter
			/// </summary>
			private string description;

			/// <summary>
			/// Initializes a new TypeParameterElement
			/// </summary>
			/// <param name="parameterName">The name of the type parameter</param>
			/// <param name="parameterDescription">The description of the type parameter</param>
			internal TypeParameterElement(string typeParameterName, string typeParameterDescription) {
				if (!string.IsNullOrEmpty(typeParameterName)) {
					name = typeParameterName.Trim();
				} else {
					name = typeParameterName;
				}
				if (!string.IsNullOrEmpty(typeParameterDescription)) {
					description = typeParameterDescription.Trim();
				} else {
					description = typeParameterDescription;
				}
			}

			/// <summary>
			/// Gets or sets the name of the type parameter
			/// </summary>
			internal string TypeParameterName {
				get {
					return name;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						name = value.Trim();
					} else {
						name = value;
					}
				}
			}

			/// <summary>
			/// Gets or sets the description of the type parameter
			/// </summary>
			internal string Description {
				get {
					return description;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						description = value.Trim();
					} else {
						description = value;
					}
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				TypeParameterElement typeParameterElement = (TypeParameterElement)otherObject;
				if (name != typeParameterElement.TypeParameterName) {
					return false;
				}
				return description == typeParameterElement.Description;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(TypeParameterElement element1, TypeParameterElement element2) {
				return element1.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(TypeParameterElement element1, TypeParameterElement element2) {
				return !element1.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return name.GetHashCode() ^ description.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<typeparam name=\"{2}\">{3}</typeparam>",
					Environment.NewLine,
					linePrefix,
					name,
					description.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

		/// <summary>
		/// A list of <see cref="Simplabs.NetDocumentor.Documentation.TypeParameterElement"/> instances
		/// </summary>
		internal class TypeParameterElementList : IEnumerable<TypeParameterElement> {

			/// <summary>
			/// The actual list that stores the elements
			/// </summary>
			private List<TypeParameterElement> list;

			/// <summary>
			/// Initializes a new TypeParameterElement
			/// </summary>
			internal TypeParameterElementList() {
				list = new List<TypeParameterElement>();
			}

			/// <summary>
			/// Gets the type parameter at index <paramref name="i"/>
			/// </summary>
			/// <param name="i">The index to get the type parameter from</param>
			/// <returns>The type parameter at index <paramref name="i"/></returns>
			internal TypeParameterElement this[int i] {
				get {
					return list[i];
				}
			}

			/// <summary>
			/// Gets the parameter with name <paramref name="parameterName"/>
			/// </summary>
			/// <param name="parameterName">The name of the parameter to get</param>
			/// <returns>The parameter with name <paramref name="parameterName"/> or nulll if no such parameter exists</returns>
			internal TypeParameterElement? this[string typeParameterName] {
				get {
					foreach (TypeParameterElement typeParameter in list) {
						if (typeParameter.TypeParameterName == typeParameterName) {
							return typeParameter;
						}
					}
					return null;
				}
			}

			/// <summary>
			/// Adds a type parameter to the list
			/// </summary>
			/// <param name="typeParameter">The type parameter to add</param>
			internal void Add(TypeParameterElement typeParameter) {
				foreach (TypeParameterElement element in list) {
					if (element.TypeParameterName == typeParameter.TypeParameterName) {
						return;
					}
				}
				list.Add(typeParameter);
			}

			/// <summary>
			/// Removes all Type Parameter from the list
			/// </summary>
			internal void Clear() {
				list.Clear();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			IEnumerator<TypeParameterElement> IEnumerable<TypeParameterElement>.GetEnumerator() {
				return list.GetEnumerator();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
				return ((IEnumerable)list).GetEnumerator();
			}

			/// <summary>
			/// Gets the string representation of the type parameters in the list
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the type parameters in the list</returns>
			internal string ToString(string linePrefix) {
				string comment = string.Empty;
				foreach (TypeParameterElement typeParameter in list) {
					comment += typeParameter.ToString(linePrefix);
				}
				return comment;
			}

		}

		/// <summary>
		/// An exception element
		/// </summary>
		internal struct ExceptionElement {

			/// <summary>
			/// The fully qualified type name of the exception
			/// </summary>
			private string fullTypeName;

			/// <summary>
			/// The description of the exception
			/// </summary>
			private string description;

			/// <summary>
			/// Initializes a new ExceptionElement
			/// </summary>
			/// <param name="typeName">The fully qualified type name of the exception</param>
			/// <param name="exceptionDescription">The description of the exception</param>
			internal ExceptionElement(string typeName, string exceptionDescription) {
				if (!string.IsNullOrEmpty(typeName)) {
					fullTypeName = typeName.Trim().Replace("<", "{").Replace(">", "}");
				} else {
					fullTypeName = string.Empty;
				}
				if (!string.IsNullOrEmpty(exceptionDescription)) {
					description = exceptionDescription.Trim();
				} else {
					description = exceptionDescription;
				}
			}

			/// <summary>
			/// Gets or sets the fully qualified type name of the exception
			/// </summary>
			internal string FullTypeName {
				get {
					return fullTypeName;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						fullTypeName = value.Trim();
					} else {
						fullTypeName = value;
					}
				}
			}

			/// <summary>
			/// Gets or sets the description of the exception
			/// </summary>
			internal string Description {
				get {
					return description;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						description = value.Trim();
					} else {
						description = value;
					}
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				ExceptionElement exceptionElement = (ExceptionElement)otherObject;
				if (fullTypeName != exceptionElement.FullTypeName) {
					return false;
				}
				return description == exceptionElement.Description;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(ExceptionElement element1, ExceptionElement element2) {
				return element1.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(ExceptionElement element1, ExceptionElement element2) {
				return !element1.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return fullTypeName.GetHashCode() ^ description.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<exception cref=\"{2}\">{3}</exception>",
					Environment.NewLine,
					linePrefix,
					fullTypeName,
					description.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

		/// <summary>
		/// A list of <see cref="Simplabs.NetDocumentor.Documentation.ExceptionElement"/> instances
		/// </summary>
		internal class ExceptionElementList : IEnumerable<ExceptionElement> {

			/// <summary>
			/// The actual list that stores the elements
			/// </summary>
			private List<ExceptionElement> list;

			/// <summary>
			/// Initializes a new ExceptionParameterElementList
			/// </summary>
			internal ExceptionElementList() {
				list = new List<ExceptionElement>();
			}

			/// <summary>
			/// Gets the exception at index <paramref name="i"/>
			/// </summary>
			/// <param name="i">The index to get the exception from</param>
			/// <returns>The exception at index <paramref name="i"/></returns>
			internal ExceptionElement this[int i] {
				get {
					return list[i];
				}
			}

			/// <summary>
			/// Adds an exception to the list
			/// </summary>
			/// <param name="exception">The exception to add</param>
			internal void Add(ExceptionElement exception) {
				foreach (ExceptionElement element in list) {
					if (element.FullTypeName == exception.FullTypeName) {
						return;
					}
				}
				list.Add(exception);
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			IEnumerator<ExceptionElement> IEnumerable<ExceptionElement>.GetEnumerator() {
				return list.GetEnumerator();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
				return ((IEnumerable)list).GetEnumerator();
			}

			/// <summary>
			/// Gets the string representation of the exceptions in the list
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the exceptions in the list</returns>
			internal string ToString(string linePrefix) {
				string comment = string.Empty;
				foreach (ExceptionElement exception in list) {
					comment += exception.ToString(linePrefix);
				}
				return comment;
			}

		}

		/// <summary>
		/// A reference element
		/// </summary>
		internal struct ReferenceElement {

			/// <summary>
			/// The fully qualified type name of the referenced type
			/// </summary>
			private string fullTypeName;

			/// <summary>
			/// Initializes a new ReferenceElement
			/// </summary>
			/// <param name="typeName">The fully qualified type name of the referenced type</param>
			internal ReferenceElement(string typeName) {
				if (!string.IsNullOrEmpty(typeName)) {
					fullTypeName = typeName.Trim();
				} else {
					fullTypeName = string.Empty;
				}
			}

			/// <summary>
			/// Gets or sets the fully qualified type name of the referenced type
			/// </summary>
			internal string FullTypeName {
				get {
					return fullTypeName;
				}
				set {
					if (!string.IsNullOrEmpty(value)) {
						fullTypeName = value.Trim();
					} else {
						fullTypeName = value;
					}
				}
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				ReferenceElement referenceElement = (ReferenceElement)otherObject;
				return fullTypeName == referenceElement.FullTypeName;
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(ReferenceElement element1, ReferenceElement element2) {
				return element1.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(ReferenceElement element1, ReferenceElement element2) {
				return !element1.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return fullTypeName.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}<seealso cref=\"{2}\"/>",
					Environment.NewLine,
					linePrefix,
					fullTypeName
				);
			}

		}

		/// <summary>
		/// A list of <see cref="Simplabs.NetDocumentor.Documentation.ReferenceElement"/> instances
		/// </summary>
		internal class ReferenceElementList : IEnumerable<ReferenceElement> {

			/// <summary>
			/// The actual list that stores the elements
			/// </summary>
			private List<ReferenceElement> list;

			/// <summary>
			/// Initializes a new ReferenceElementList
			/// </summary>
			internal ReferenceElementList() {
				list = new List<ReferenceElement>();
			}

			/// <summary>
			/// Gets the reference at index <paramref name="i"/>
			/// </summary>
			/// <param name="i">The index to get the > from</param>
			/// <returns>The > at index <paramref name="i"/></returns>
			internal ReferenceElement this[int i] {
				get {
					return list[i];
				}
			}

			/// <summary>
			/// Adds a reference to the list
			/// </summary>
			/// <param name="reference">The reference to add</param>
			internal void Add(ReferenceElement reference) {
				foreach (ReferenceElement element in list) {
					if (element.FullTypeName == reference.FullTypeName) {
						return;
					}
				}
				list.Add(reference);
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			IEnumerator<ReferenceElement> IEnumerable<ReferenceElement>.GetEnumerator() {
				return list.GetEnumerator();
			}

			/// <summary>
			/// Gets an enumerator for the list
			/// </summary>
			/// <returns>An enumerator for the list</returns>
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
				return ((IEnumerable)list).GetEnumerator();
			}

			/// <summary>
			/// Gets the string representation of the references in the list
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the references in the list</returns>
			internal string ToString(string linePrefix) {
				string comment = string.Empty;
				foreach (ReferenceElement reference in list) {
					comment += reference.ToString(linePrefix);
				}
				return comment;
			}

		}

		/// <summary>
		/// A documentation element that is not supported by .Net Documentor
		/// </summary>
		internal struct UnsupportedElement {

			/// <summary>
			/// The <see cref="System.Xml.XmlNode"/> that represents the element
			/// </summary>
			private XmlNode node;

			/// <summary>
			/// Initializes a new UnsupportedElement
			/// </summary>
			/// <param name="elementNode">The <see cref="System.Xml.XmlNode"/> that represents the element</param>
			internal UnsupportedElement(XmlNode elementNode) {
				node = elementNode;
			}

			/// <summary>
			/// Gets whether <paramref name="otherObject"/> equals the current
			/// </summary>
			/// <param name="otherObject">The object to compare the current to</param>
			/// <returns>True if <paramref name="otherObject"/> equals the current object, false otherwise</returns>
			public override bool Equals(object otherObject) {
				if (otherObject == null) {
					return false;
				}
				if (GetType() != otherObject.GetType()) {
					return false;
				}
				UnsupportedElement element = (UnsupportedElement)otherObject;
				return ToString(string.Empty) == element.ToString(string.Empty);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are equal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are equal, false otherwise</returns>
			public static bool operator ==(UnsupportedElement element1, UnsupportedElement element2) {
				return element1.Equals(element2);
			}

			/// <summary>
			/// Gets whether <paramref name="element1"/> and <paramref name="element2"/> are unequal
			/// </summary>
			/// <param name="element1">The first element to compare</param>
			/// <param name="element2">The second element to compare</param>
			/// <returns>True if <paramref name="element1"/> and <paramref name="element2"/> are unequal, false otherwise</returns>
			public static bool operator !=(UnsupportedElement element1, UnsupportedElement element2) {
				return !element1.Equals(element2);
			}

			/// <summary>
			/// Gets a hash code for the current instance
			/// </summary>
			/// <returns>A hash code for the current instance</returns>
			public override int GetHashCode() {
				return node.GetHashCode();
			}

			/// <summary>
			/// Gets the string representation of the documentation element
			/// </summary>
			/// <param name="linePrefix">The prefix to prepend to any line of the string representation (e.g. <code>///</code> for languages that don't add this prefix automaticallly)</param>
			/// <returns>The string representation of the documentation element</returns>
			internal string ToString(string linePrefix) {
				return string.Format(
					CultureInfo.CurrentCulture,
					"{0}{1}{2}",
					Environment.NewLine,
					linePrefix,
					node.OuterXml.Replace(Environment.NewLine, Environment.NewLine + linePrefix)
				);
			}

		}

	}


}