using System;
using System.Collections.Generic;
using System.Text;
using EnvDTE;
using EnvDTE90;
//using EnvDTE80;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Collections;
using System.Runtime.InteropServices;
using EnvDTE80;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// The base class for all documentor that document managed code
	/// </summary>
	internal abstract class ManagedCodeDocumentor {

		/// <summary>
		/// Finds a code element matching the signature of <paramref name="element"/> in type <paramref name="theType"/>
		/// </summary>
		/// <param name="element">The code element to find a matching code element for</param>
		/// <param name="theType">The type to find the matching code element in (all base types of the type are also included in the search)</param>
		/// <param name="searchedElementKind">The kind of element that is searched for</param>
		/// <param name="elementPrototype">The prototype of the element that is searched for</param>
		/// <returns>The code element matching the signature of <paramref name="element"/> or null if no such element exists</returns>
		private CodeElement2 FindMatchingBaseTypeMember(CodeElement2 element, CodeType theType, vsCMElement searchedElementKind, string elementPrototype) {
			if (element.Kind != vsCMElement.vsCMElementFunction && element.Kind != vsCMElement.vsCMElementEvent && element.Kind != vsCMElement.vsCMElementProperty) {
				return null;
			}
			foreach (CodeElement baseTypeElement in theType.Members) {
				if (baseTypeElement.Kind == searchedElementKind) {
					string baseElementPrototype = string.Empty;
					if (baseTypeElement.Kind == vsCMElement.vsCMElementFunction) {
						baseElementPrototype = ((CodeFunction2)baseTypeElement).get_Prototype(
							(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
						);
					} else if (baseTypeElement.Kind == vsCMElement.vsCMElementEvent) {
						baseElementPrototype = ((CodeEvent)baseTypeElement).Type.AsFullName + ":" + ((CodeEvent)baseTypeElement).Name;
					} else if (baseTypeElement.Kind == vsCMElement.vsCMElementProperty) {
						baseElementPrototype = ((CodeProperty)baseTypeElement).get_Prototype(
							(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
						);
					}
					if (!string.IsNullOrEmpty(baseElementPrototype) && baseElementPrototype.Equals(elementPrototype)) {
						return (CodeElement2)baseTypeElement;
					}
				}
			}
			foreach (CodeType superBaseType in theType.Bases) {
				if (superBaseType.IsCodeType) {
					CodeElement2 upperLevelmatch = FindMatchingBaseTypeMember(element, superBaseType, searchedElementKind, elementPrototype);
					if (upperLevelmatch != null) {
						return upperLevelmatch;
					}
				}
			}
			if (theType.Kind == vsCMElement.vsCMElementClass) {
				foreach (CodeType theInterface in ((CodeClass2)theType).ImplementedInterfaces) {
					CodeElement2 upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface, searchedElementKind, elementPrototype);
					if (upperLevelmatch != null) {
						return upperLevelmatch;
					}
				}
			} else if (theType.Kind == vsCMElement.vsCMElementStruct) {
				foreach (CodeType theInterface in ((CodeStruct2)theType).ImplementedInterfaces) {
					CodeElement2 upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface, searchedElementKind, elementPrototype);
					if (upperLevelmatch != null) {
						return upperLevelmatch;
					}
				}
			} else if (theType.Kind == vsCMElement.vsCMElementInterface) {
				foreach (CodeType theInterface in ((CodeInterface2)theType).Bases) {
					CodeElement2 upperLevelmatch = FindMatchingBaseTypeMember(element, theInterface, searchedElementKind, elementPrototype);
					if (upperLevelmatch != null) {
						return upperLevelmatch;
					}
				}
			}
			return null;
		}

		/// <summary>
		/// Gets a documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.
		/// </summary>
		/// <param name="element">The code element to get the documentation comment for</param>
		/// <returns>A documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.</returns>
		/// <remarks>If the code element is an overridden method, a method giding another from the base type or an element implementing an interface and does not have a documentation comment itself, the documentation comment of the matching base type's member will be returned if available</remarks>
		protected DocumentationComment GetDocumentationComment(CodeElement2 element) {
			return GetDocumentationComment(element, false);
		}

		/// <summary>
		/// Gets a documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.
		/// </summary>
		/// <param name="element">The code element to get the documentation comment for</param>
		/// <param name="resuseBaseTypeDocumentation">Specifies whether base type documentations are to be reused for override, hides or interface implementations</param>
		/// <returns>A documentation comment for code element <paramref name="element"/>. This is either a documentation comment based on the existing documentation of the element, or a new one if no documentation exists for the element.</returns>
		/// <remarks>If the code element is an overridden method, a method giding another from the base type or an element implementing an interface and does not have a documentation comment itself, the documentation comment of the matching base type's member will be returned if available</remarks>
		protected DocumentationComment GetDocumentationComment(CodeElement2 element, bool resuseBaseTypeDocumentation) {
			if (SourceCodeServices.HasDocumentationComment(element)) {
				return SourceCodeServices.ParseDocumentationComment(element, string.Empty);
			}
			if (resuseBaseTypeDocumentation && (
				element.Kind == vsCMElement.vsCMElementFunction || element.Kind == vsCMElement.vsCMElementEvent || element.Kind == vsCMElement.vsCMElementProperty
			)) {
				CodeElement2 matchingBaseTypeElement = null;
				CodeType containingType = null;
				string elementPrototype = string.Empty;
				vsCMElement searchedElementKind = vsCMElement.vsCMElementOther;
				if (element.Kind == vsCMElement.vsCMElementFunction) {
					try {
						containingType = (CodeType)(((CodeFunction2)element).Parent);
					} catch (Exception) {
						return new DocumentationComment();
					}
					searchedElementKind = vsCMElement.vsCMElementFunction;
					elementPrototype = ((CodeFunction2)element).get_Prototype(
						(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
					);
				} else if (element.Kind == vsCMElement.vsCMElementEvent) {
					try{
						containingType = (CodeType)(((CodeEvent)element).Parent);
					} catch (Exception) {
						return new DocumentationComment();
					}
					searchedElementKind = vsCMElement.vsCMElementEvent;
					elementPrototype = ((CodeEvent)element).Type.AsFullName + ":" + ((CodeEvent)element).Name;
				} else if (element.Kind == vsCMElement.vsCMElementProperty) {
					try {
						containingType = (CodeType)(((CodeProperty2)element).Parent2);
					} catch (Exception) {
						return new DocumentationComment();
					}
					searchedElementKind = vsCMElement.vsCMElementProperty;
					elementPrototype = ((CodeProperty)element).get_Prototype(
						(int)(vsCMPrototype.vsCMPrototypeType | vsCMPrototype.vsCMPrototypeParamTypes)
					);
				}
				if (element.Name.Contains(".")) {
					elementPrototype = Regex.Replace(elementPrototype, @"\s[^\.]+\.", " ", RegexOptions.IgnoreCase | RegexOptions.Compiled | RegexOptions.Singleline);
				}
				foreach (CodeType baseType in containingType.Bases) {
					matchingBaseTypeElement = FindMatchingBaseTypeMember(element, baseType, searchedElementKind, elementPrototype);
					if (matchingBaseTypeElement != null) {
						return SourceCodeServices.ParseDocumentationComment(matchingBaseTypeElement, string.Empty);
					}
				}
				CodeElements implementedInterfaces = null;
				if (containingType.Kind == vsCMElement.vsCMElementClass) {
					implementedInterfaces = ((CodeClass2)containingType).ImplementedInterfaces;
				} else if (containingType.Kind == vsCMElement.vsCMElementStruct) {
					implementedInterfaces = ((CodeStruct2)containingType).ImplementedInterfaces;
				}
				if (implementedInterfaces != null) {
					foreach (CodeType implementedInterface in implementedInterfaces) {
						matchingBaseTypeElement = FindMatchingBaseTypeMember(element, implementedInterface, searchedElementKind, elementPrototype);
						if (matchingBaseTypeElement != null) {
							return SourceCodeServices.ParseDocumentationComment(matchingBaseTypeElement, string.Empty);
						}
					}
				}
			}
			return new DocumentationComment();
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as references to types
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		protected void DocumentAsTypeReferences(CodeElements elements, DocumentationComment comment) {
			foreach (CodeElement referencedType in elements) {
				if (
					!referencedType.FullName.Equals(typeof(System.Object).FullName, StringComparison.InvariantCultureIgnoreCase)
					&& !referencedType.FullName.Equals(typeof(System.ValueType).FullName, StringComparison.InvariantCultureIgnoreCase)
				) {
					DocumentationComment.ReferenceElement referenceElement = new DocumentationComment.ReferenceElement("T:" + referencedType.FullName);
					comment.References.Add(referenceElement);
				}
			}
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as type parameters
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		protected static void DocumentAsTypeParameters(List<string> elements, DocumentationComment comment) {
			List<DocumentationComment.TypeParameterElement> typeParameters = new List<DocumentationComment.TypeParameterElement>();
			foreach (string typeParameter in elements) {
				if (comment.TypeParameters[typeParameter] != null) {
					typeParameters.Add(comment.TypeParameters[typeParameter].Value);
					continue;
				}
				typeParameters.Add(new DocumentationComment.TypeParameterElement(typeParameter, string.Empty));
			}
			comment.TypeParameters.Clear();
			foreach (DocumentationComment.TypeParameterElement typeParameter in typeParameters) {
				comment.TypeParameters.Add(typeParameter);
			}
		}

		/// <summary>
		/// Documents <paramref name="elements"/> as parameters
		/// </summary>
		/// <param name="elements">The elements to document as references</param>
		/// <param name="comment">The comment to write the documentation to</param>
		protected static void DocumentAsParameters(IEnumerable elements, DocumentationComment comment) {
			List<DocumentationComment.ParameterElement> parameters = new List<DocumentationComment.ParameterElement>();
			foreach (CodeElement2 parameter in elements) {
				if (comment.Parameters[parameter.Name] != null) {
					parameters.Add(comment.Parameters[parameter.Name].Value);
					continue;
				}
				parameters.Add(new DocumentationComment.ParameterElement(parameter.Name, string.Empty));
			}
			comment.Parameters.Clear();
			foreach (DocumentationComment.ParameterElement parameter in parameters) {
				comment.Parameters.Add(parameter);
			}
		}

	}

}