using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using EnvDTE;
using Microsoft.VisualStudio.VCCodeModel;
using System.IO;
using System.Diagnostics;
using EnvDTE80;
using System.Runtime.InteropServices;
using System.Xml;
using System.Globalization;

namespace Simplabs.NetDocumentor.Documentation {

	/// <summary>
	/// Implements some helpers methods for modifications on the source code buffer
	/// </summary>
	internal static class SourceCodeServices {

		/// <summary>
		/// Gets an EditPoint at the start point of <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to get the EditPoint for</param>
		/// <param name="elementPortion">Specifies the portion of the element to get the edit point at (either declearation or definition)</param>
		/// <returns>An EditPoint at the position of declaration of <paramref name="element"/></returns>
		/// <remarks>In case that the element has attributes assigned to it and <paramref name="elementPortion"/> is <see cref="Microsoft.VisualStudio.VCCodeModel.vsCMWhere.vsCMWhereDeclaration"/>, the point at the beginning of the attribute statements is returned.</remarks>
		private static EditPoint GetRealStartPoint(VCCodeElement element, vsCMWhere elementPortion) {
			try {
				TextPoint sourceCodePoint;
				sourceCodePoint = element.get_StartPointOf(vsCMPart.vsCMPartWholeWithAttributes, elementPortion);
				return sourceCodePoint.CreateEditPoint();
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			}
		}

		/// <summary>
		/// Gets whether <paramref name="comment"/> represents an empty documentation comment or not
		/// </summary>
		/// <param name="commment">The comment to check whether it is empty or not</param>
		/// <returns>True if <paramref name="comment"/> represents an empty doocumentation comment, false otherwise</returns>
		private static bool IsEmptyComment(string commment) {
			if (string.IsNullOrEmpty(commment)) {
				return true;
			}
			if (Regex.Match(commment.Trim(), @"<doc>\s*<\/doc>").Success) {
				return true;
			}
			return false;
		}

		/// <summary>
		/// Inserts the XML comment for <paramref name="element"/> and shows the inserted comment in the code editor (if specified via <param name="showInsertionPoint"/>)
		/// </summary>
		/// <param name="element">The element to add the XML comment for</param>
		/// <param name="documentation">The documentation text to add</param>
		/// <param name="elementPortion">Specifies the portion of the element to insert the XML comment at</param>
		/// <param name="insertionPoint">The point in the source code to insert the XML comment at</param>
		private static void InsertDocumentation(VCCodeElement element, string documentation, vsCMWhere elementPortion, EditPoint insertionPoint) {
			bool wasCommentRemoved = RemoveExistingComment(element, elementPortion);
			if (wasCommentRemoved) {
				insertionPoint.LineUp(1);
				insertionPoint.EndOfLine();
				insertionPoint.Insert(documentation);
			} else {
				insertionPoint.Insert(documentation + Environment.NewLine);
			}
			int commentLines = documentation.Split(new string[] { Environment.NewLine },	StringSplitOptions.None).Length;
			EditPoint endPoint = insertionPoint.CreateEditPoint();
			insertionPoint.LineUp(commentLines);
			insertionPoint.SmartFormat(endPoint);
		}

		/// <summary>
		/// Shows the declaration position of an element in the code editor
		/// </summary>
		/// <param name="element">The element to show</param>
		/// <param name="elementPortion">Specifies the portion of the element to show</param>
		internal static void ShowDocumentation(VCCodeElement element, vsCMWhere elementPortion) {
			try {
				EditPoint sourceCodePoint = SourceCodeServices.GetRealStartPoint(element, elementPortion);
				element.DTE.OpenFile(EnvDTE.Constants.vsViewKindCode, Path.Combine(sourceCodePoint.Parent.Parent.Path, sourceCodePoint.Parent.Parent.Name));
				sourceCodePoint.Parent.Parent.Activate();
				sourceCodePoint.TryToShow(vsPaneShowHow.vsPaneShowCentered, 0);
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			}
		}

		/// <summary>
		/// Removes any existing comments from code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to remove existing comments from</param>
		/// <param name="elementPortion">Specifies the portion of the element to remove existing comments from</param>
		/// <returns>True if any coments hhave been removed, false otherwise</returns>
		private static bool RemoveExistingComment(VCCodeElement element, vsCMWhere elementPortion) {
			if (HasComment(element, elementPortion)) {
				EditPoint sourceCodePoint = SourceCodeServices.GetRealStartPoint(element, elementPortion);
				sourceCodePoint.LineUp(1);
				sourceCodePoint.StartOfLine();
				string lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
				while (lineText.StartsWith("//")) {
					sourceCodePoint.Delete(sourceCodePoint.LineLength);
					sourceCodePoint.LineUp(1);
					lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
					if (lineText.StartsWith("//")) {
						sourceCodePoint.LineDown(1);
						sourceCodePoint.DeleteWhitespace(vsWhitespaceOptions.vsWhitespaceOptionsVertical);
						sourceCodePoint.LineUp(1);
					}
				}
				if (lineText.EndsWith("*/")) {
					while (!lineText.Contains("/*")) {
						sourceCodePoint.Delete(sourceCodePoint.LineLength);
						sourceCodePoint.DeleteWhitespace(vsWhitespaceOptions.vsWhitespaceOptionsVertical);
						sourceCodePoint.LineUp(1);
						lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
					}
					sourceCodePoint.MoveToLineAndOffset(sourceCodePoint.Line, lineText.LastIndexOf("/*") + 1);
					sourceCodePoint.Delete(sourceCodePoint.LineLength - lineText.LastIndexOf("/*"));
				}
				return true;
			}
			return false;
		}

		/// <summary>
		/// Gets the code element from a collection of <paramref name="codeElements"/> that is located at a specified <paramref name="textPoint"/>
		/// </summary>
		/// <param name="codeElements">The collection of code elements to find the code element that is located at the spcified <paramref name="textPoint"/> in</param>
		/// <param name="textPoint">The position of the code element that is searched for</param>
		/// <param name="bestMatch">The best matching code element that has been fond so far (for recursive calls, set to defalt retrn value on first call)</param>
		/// <returns>The code element that is located at the specified <paramref name="textPoint"/> or the default value <paramref name="bestMatch"/> if no sch element exists</returns>
		private static CodeElement2 GetCodeElementAtTextPoint(CodeElements codeElements, TextPoint textPoint, CodeElement2 bestMatch) {
			foreach (CodeElement2 element in codeElements) {
				try {
					if (!element.StartPoint.GreaterThan(textPoint) && !element.EndPoint.LessThan(textPoint)) {
						bestMatch = GetCodeElementAtTextPoint(element.Children, textPoint, element);
					}
				} catch (Exception) {
				}
			}
			return bestMatch;
		}

		/// <summary>
		/// Gets the text of the documentation comment for code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The code element to get the text of the documentation comment for</param>
		/// <returns>The text of the documentation comment for code element <paramref name="element"/></returns>
		private static string GetDocumentationCommentText(CodeElement2 element) {
			string documentationCommentText = string.Empty;
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
					documentationCommentText = ((CodeClass)element).DocComment;
					break;
				case vsCMElement.vsCMElementFunction:
					documentationCommentText = ((CodeFunction)element).DocComment;
					break;
				case vsCMElement.vsCMElementProperty:
					documentationCommentText = ((CodeProperty)element).DocComment;
					break;
				case vsCMElement.vsCMElementVariable:
					documentationCommentText = ((CodeVariable)element).DocComment;
					break;
				case vsCMElement.vsCMElementEvent:
					documentationCommentText = ((CodeEvent)element).DocComment;
					break;
				case vsCMElement.vsCMElementInterface:
					documentationCommentText = ((CodeInterface)element).DocComment;
					break;
				case vsCMElement.vsCMElementStruct:
					documentationCommentText = ((CodeStruct)element).DocComment;
					break;
				case vsCMElement.vsCMElementEnum:
					documentationCommentText = ((CodeEnum)element).DocComment;
					break;
				case vsCMElement.vsCMElementDelegate:
					documentationCommentText = ((CodeDelegate)element).DocComment;
					break;
				default:
					break;
			}
			return documentationCommentText;
		}

		/// <summary>
		/// Gets the text of the documentation comment for code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The code element to get the text of the documentation comment for</param>
		/// <param name="elementPortion">Specifies the portion of the element to get the documentation comment from</param>
		/// <returns>The text of the documentation comment for code element <paramref name="element"/></returns>
		private static string GetDocumentationCommentText(VCCodeElement element, vsCMWhere elementPortion) {
			EditPoint sourceCodePoint = SourceCodeServices.GetRealStartPoint(element, elementPortion);
			sourceCodePoint.LineUp(1);
			sourceCodePoint.StartOfLine();
			StringBuilder documentationCommentText = new StringBuilder();
			string lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
			while (lineText.StartsWith("///")) {
				documentationCommentText.Insert(0, lineText.Substring(3) + Environment.NewLine);
				sourceCodePoint.LineUp(1);
				lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
			}
			return documentationCommentText.ToString();
		}

		/// <summary>
		/// Gets the text of the documentation comment for external code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The external code element to get the text of the documentation comment for</param>
		/// <returns>The text of the documentation comment for external code element <paramref name="element"/></returns>
		private static string GetDocumentationCommentTextFromExternal(VCCodeElement element) {
			string documentationCommentText = string.Empty;
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
					documentationCommentText = ((VCCodeClass)element).DocComment;
					break;
				case vsCMElement.vsCMElementFunction:
					documentationCommentText = ((VCCodeFunction)element).DocComment;
					break;
				case vsCMElement.vsCMElementProperty:
					documentationCommentText = ((VCCodeProperty)element).DocComment;
					break;
				case vsCMElement.vsCMElementVariable:
					documentationCommentText = ((VCCodeVariable)element).DocComment;
					break;
				case vsCMElement.vsCMElementInterface:
					documentationCommentText = ((VCCodeInterface)element).DocComment;
					break;
				case vsCMElement.vsCMElementStruct:
					documentationCommentText = ((VCCodeStruct)element).DocComment;
					break;
				case vsCMElement.vsCMElementEnum:
					documentationCommentText = ((VCCodeEnum)element).DocComment;
					break;
				case vsCMElement.vsCMElementDelegate:
					documentationCommentText = ((VCCodeDelegate)element).DocComment;
					break;
				default:
					break;
			}
			return documentationCommentText;
		}

		/// <summary>
		/// Builds a documentation comment from its string representation
		/// </summary>
		/// <param name="documentationCommentText">The string representation of the documentation comment</param>
		/// <param name="linePrefix">The line prefix to use for the documentation comment (e.g. <code>///</code>)</param>
		/// <returns>A <see cref="Simplabs.NetDocumentor.Documentation.DocumentationComment"/> retrieved from the string representation <paramref name="documentationCommentText"/></returns>
		internal static DocumentationComment BuildDocumentationComment(string documentationCommentText, string linePrefix) {
			try {
				if (!documentationCommentText.Trim().StartsWith("<doc>")) {
					documentationCommentText = "<doc>" + documentationCommentText;
				}
				if (!documentationCommentText.Trim().EndsWith("</doc>")) {
					documentationCommentText = documentationCommentText + "</doc>";
				}
				DocumentationComment documentationComment = new DocumentationComment(linePrefix);
				XmlDocument xmlData = new XmlDocument();
				xmlData.LoadXml(documentationCommentText);
				XmlNode summaryNode = xmlData.SelectSingleNode("/doc/summary");
				if (summaryNode != null) {
					documentationComment.Summary = new DocumentationComment.SummaryElement(summaryNode.InnerXml);
					xmlData.DocumentElement.RemoveChild(summaryNode);
				}
				XmlNode returnsNode = xmlData.SelectSingleNode("/doc/returns");
				if (returnsNode != null) {
					documentationComment.Returns = new DocumentationComment.ReturnsElement(returnsNode.InnerXml);
					xmlData.DocumentElement.RemoveChild(returnsNode);
				}
				XmlNode valueNode = xmlData.SelectSingleNode("/doc/value");
				if (valueNode != null) {
					documentationComment.Value = new DocumentationComment.ValueElement(valueNode.InnerXml);
					xmlData.DocumentElement.RemoveChild(valueNode);
				}
				XmlNodeList parameterNodes = xmlData.SelectNodes("/doc/param");
				foreach (XmlNode parameterNode in parameterNodes) {
					documentationComment.Parameters.Add(new DocumentationComment.ParameterElement(parameterNode.Attributes["name"] != null ? parameterNode.Attributes["name"].Value : string.Empty, parameterNode.InnerXml));
					xmlData.DocumentElement.RemoveChild(parameterNode);
				}
				XmlNodeList exceptionNodes = xmlData.SelectNodes("/doc/exception");
				foreach (XmlNode exceptionNode in exceptionNodes) {
					documentationComment.Exceptions.Add(new DocumentationComment.ExceptionElement(exceptionNode.Attributes["cref"] != null ? exceptionNode.Attributes["cref"].Value : string.Empty, exceptionNode.InnerXml));
					xmlData.DocumentElement.RemoveChild(exceptionNode);
				}
				XmlNodeList referenceNodes = xmlData.SelectNodes("/doc/seealso");
				foreach (XmlNode referenceNode in referenceNodes) {
					documentationComment.References.Add(new DocumentationComment.ReferenceElement(referenceNode.Attributes["cref"] != null ? referenceNode.Attributes["cref"].Value : string.Empty));
					xmlData.DocumentElement.RemoveChild(referenceNode);
				}
				XmlNodeList typeParameterNodes = xmlData.SelectNodes("/doc/typeparam");
				foreach (XmlNode typeParameterNode in typeParameterNodes) {
					documentationComment.TypeParameters.Add(new DocumentationComment.TypeParameterElement(typeParameterNode.Attributes["name"] != null ? typeParameterNode.Attributes["name"].Value : string.Empty, typeParameterNode.InnerXml));
					xmlData.DocumentElement.RemoveChild(typeParameterNode);
				}
				XmlNodeList unsupportedNodes = xmlData.DocumentElement.ChildNodes;
				documentationComment.SetUnsupportedElements(unsupportedNodes);
				return documentationComment;
			} catch (Exception) {
				return new DocumentationComment(linePrefix);
			}
		}

		/// <summary>
		/// Gets the code element from a collection of <paramref name="codeElements"/> that is located at a specified <paramref name="textPoint"/>
		/// </summary>
		/// <param name="codeElements">The code model to find the element that is located at the specified <paramref name="textPoint"/> in</param>
		/// <param name="textPoint">The position of the code element that is searched for</param>
		/// <returns>The code element that is located at the specified <paramref name="textPoint"/> or null if no sch element exists</returns>
		internal static CodeElement2 GetCodeElementAtTextPoint(FileCodeModel2 codeModel, TextPoint textPoint) {
			if (codeModel.Language == CodeModelLanguageConstants.vsCMLanguageMC || codeModel.Language == CodeModelLanguageConstants.vsCMLanguageVC) {
				return (CodeElement2)codeModel.CodeElementFromPoint(
					textPoint,
					vsCMElement.vsCMElementFunction | vsCMElement.vsCMElementProperty | vsCMElement.vsCMElementVariable | vsCMElement.vsCMElementEvent | vsCMElement.vsCMElementClass | vsCMElement.vsCMElementInterface | vsCMElement.vsCMElementStruct | vsCMElement.vsCMElementEnum
				);
			}
			return GetCodeElementAtTextPoint(codeModel.CodeElements, textPoint, null);
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> has a comment or not
		/// </summary>
		/// <param name="element">The element to check for whether it has a comment or not</param>
		/// <returns>True if <paramref name="element"/> has a comment, false otherwise</returns>
		internal static bool HasComment(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
					return !IsEmptyComment(((CodeClass)element).DocComment) || !IsEmptyComment(((CodeClass)element).Comment);
				case vsCMElement.vsCMElementFunction:
					return !IsEmptyComment(((CodeFunction)element).DocComment) || !IsEmptyComment(((CodeFunction)element).Comment);
				case vsCMElement.vsCMElementProperty:
					return !IsEmptyComment(((CodeProperty)element).DocComment) || !IsEmptyComment(((CodeProperty)element).Comment);
				case vsCMElement.vsCMElementVariable:
					return !IsEmptyComment(((CodeVariable)element).DocComment) || !IsEmptyComment(((CodeVariable)element).Comment);
				case vsCMElement.vsCMElementEvent:
					return !IsEmptyComment(((CodeEvent)element).DocComment) || !IsEmptyComment(((CodeEvent)element).Comment);
				case vsCMElement.vsCMElementInterface:
					return !IsEmptyComment(((CodeInterface)element).DocComment) || !IsEmptyComment(((CodeInterface)element).Comment);
				case vsCMElement.vsCMElementStruct:
					return !IsEmptyComment(((CodeStruct)element).DocComment) || !IsEmptyComment(((CodeStruct)element).Comment);
				case vsCMElement.vsCMElementEnum:
					return !IsEmptyComment(((CodeEnum)element).DocComment) || !IsEmptyComment(((CodeEnum)element).Comment);
				case vsCMElement.vsCMElementDelegate:
					return !IsEmptyComment(((CodeDelegate)element).DocComment) || !IsEmptyComment(((CodeDelegate)element).Comment);
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> has a domentation comment or not
		/// </summary>
		/// <param name="element">The element to check for whether it has a documentation comment or not</param>
		/// <returns>True if <paramref name="element"/> has a documentation comment, false otherwise</returns>
		internal static bool HasDocumentationComment(CodeElement2 element) {
			switch (element.Kind) {
				case vsCMElement.vsCMElementClass:
					return !IsEmptyComment(((CodeClass)element).DocComment);
				case vsCMElement.vsCMElementFunction:
					return !IsEmptyComment(((CodeFunction)element).DocComment);
				case vsCMElement.vsCMElementProperty:
					return !IsEmptyComment(((CodeProperty)element).DocComment);
				case vsCMElement.vsCMElementVariable:
					return !IsEmptyComment(((CodeVariable)element).DocComment);
				case vsCMElement.vsCMElementEvent:
					return !IsEmptyComment(((CodeEvent)element).DocComment);
				case vsCMElement.vsCMElementInterface:
					return !IsEmptyComment(((CodeInterface)element).DocComment);
				case vsCMElement.vsCMElementStruct:
					return !IsEmptyComment(((CodeStruct)element).DocComment);
				case vsCMElement.vsCMElementEnum:
					return !IsEmptyComment(((CodeEnum)element).DocComment);
				case vsCMElement.vsCMElementDelegate:
					return !IsEmptyComment(((CodeDelegate)element).DocComment);
				default:
					return false;
			}
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> has a comment or not
		/// </summary>
		/// <param name="element">The element to check for whether it has a comment or not</param>
		/// <param name="elementPortion">Specifies the portion of the element to check for comments at</param>
		/// <returns>True if <paramref name="element"/> has a comment, false otherwise</returns>
		internal static bool HasComment(VCCodeElement element, vsCMWhere elementPortion) {
			EditPoint sourceCodePoint = SourceCodeServices.GetRealStartPoint(element, elementPortion);
			sourceCodePoint.LineUp(1);
			sourceCodePoint.StartOfLine();
			string lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
			return lineText.StartsWith("//") || lineText.EndsWith("*/");
		}

		/// <summary>
		/// Gets whether <paramref name="element"/> has a domentation comment or not
		/// </summary>
		/// <param name="element">The element to check for whether it has a documentation comment or not</param>
		/// <param name="elementPortion">Specifies the portion of the element to check for documentation comments at</param>
		/// <returns>True if <paramref name="element"/> has a documentation comment, false otherwise</returns>
		internal static bool HasDocumentationComment(VCCodeElement element, vsCMWhere elementPortion) {
			try {
				EditPoint sourceCodePoint = SourceCodeServices.GetRealStartPoint(element, elementPortion);
				sourceCodePoint.LineUp(1);
				sourceCodePoint.StartOfLine();
				string lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
				return lineText.StartsWith("///");
			} catch (Exception) {
				return false;
			}
		}

		/// <summary>
		/// Inserts the XML comment for <paramref name="element"/> and shows the inserted comment in the code editor (if specified via <param name="showInsertionPoint"/>)
		/// </summary>
		/// <param name="element">The element to add the XML comment for</param>
		/// <param name="documentation">The documentation text to add</param>
		/// <remarks>This method also removes all existing documentation (these are all comments above the code element until the first blank or non-comment line)</remarks>
		internal static void InsertDocumentation(VCCodeElement element, string documentation) {
			bool? undoContextWasOpen = null;
			try {
				vsCMWhere elementPortion = vsCMWhere.vsCMWhereDefault;
				if (global::Simplabs.NetDocumentor.Properties.LanguageSpecificSettings.Default.CppDocumentAtDeclaration) {
					elementPortion = vsCMWhere.vsCMWhereDeclaration;
				} else if (global::Simplabs.NetDocumentor.Properties.LanguageSpecificSettings.Default.CppDocumentAtDefinition) {
					elementPortion = vsCMWhere.vsCMWhereDefinition;
				}
				undoContextWasOpen = element.DTE.UndoContext.IsOpen;
				if (!undoContextWasOpen.Value) {
					element.DTE.UndoContext.Open(
						string.Format(CultureInfo.CurrentUICulture, "Document Element '{0}'", element.DisplayName),
						false
					);
				}
				element.CodeModel.Synchronize();
				InsertDocumentation(
					element,
					documentation,
					elementPortion,
					SourceCodeServices.GetRealStartPoint(element, elementPortion)
				);
				element.CodeModel.Synchronize();
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			} finally {
				if (undoContextWasOpen.HasValue && !undoContextWasOpen.Value) {
					element.DTE.UndoContext.Close();
				}
			}
		}

		/// <summary>
		/// Inserts the XML comment for <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to add the XML comment for</param>
		/// <param name="documentation">The documentation text to add</param>
		internal static void InsertDocumentation(CodeElement2 element, string documentation) {
			bool? undoContextWasOpen = null;
			try {
				undoContextWasOpen = element.DTE.UndoContext.IsOpen;
				if (!undoContextWasOpen.Value) {
					element.DTE.UndoContext.Open(
						string.Format(CultureInfo.CurrentUICulture, "Document Element '{0}'", element.FullName),
						false
					);
				}
				RemoveExistingComment(element);
				switch (element.Kind) {
					case vsCMElement.vsCMElementClass:
						((CodeClass)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementFunction:
						((CodeFunction)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementProperty:
						((CodeProperty)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementVariable:
						((CodeVariable)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementEvent:
						((CodeEvent)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementInterface:
						((CodeInterface)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementStruct:
						((CodeStruct)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementEnum:
						((CodeEnum)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementDelegate:
						((CodeDelegate)element).DocComment = documentation;
						break;
					case vsCMElement.vsCMElementModule:
						EditPoint sourceCodePoint = element.StartPoint.CreateEditPoint();
						sourceCodePoint.LineUp(1);
						sourceCodePoint.Insert(documentation + Environment.NewLine);
						break;
					default:
						break;
				}
			} catch (Exception exception) {
				throw SourceCodeServices.PrepareDocumentationException(element, exception);
			} finally {
				if (undoContextWasOpen.HasValue && !undoContextWasOpen.Value) {
					element.DTE.UndoContext.Close();
				}
			}
		}

		/// <summary>
		/// Gets the template parameters for a code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The code element to get the template parameters for</param>
		/// <returns>The template parameters of the code element</returns>
		internal static List<string> GetTemplateParameters(CodeElement2 element) {
			List<string> templateParameters = new List<string>();
			string templateParametersList = string.Empty;
			Match match;
			switch (element.Language)
			{
				case CodeModelLanguageConstants.vsCMLanguageCSharp:
				case CodeModelLanguageConstants.vsCMLanguageMC:
				case CodeModelLanguageConstants.vsCMLanguageVC:
					match = Regex.Match(element.FullName, @"<([^>]*)>", RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.Multiline | RegexOptions.RightToLeft);
					if (match.Groups.Count > 0 && match.Groups[1].Success) {
						if (element.Language == CodeModelLanguageConstants.vsCMLanguageMC || element.Language == CodeModelLanguageConstants.vsCMLanguageVC) {
							templateParametersList = Regex.Replace(match.Groups[1].Value, @"typename\s+", string.Empty, RegexOptions.Compiled | RegexOptions.IgnoreCase);
						} else {
							templateParametersList = match.Groups[1].Value;
						}
					}
					break;
				case CodeModelLanguageConstants.vsCMLanguageVB:
					match = Regex.Match(element.FullName, @"\(\w*of([^\)]*)\)", RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.Multiline | RegexOptions.IgnoreCase | RegexOptions.RightToLeft);
					if (match.Groups.Count > 0 && match.Groups[1].Success) {
						templateParametersList = match.Groups[1].Value;
					}
					break;
				default:
					break;
			}
			foreach (string templateParameter in templateParametersList.Split(new char[] { ',' })) {
				templateParameters.Add(templateParameter.Trim());
			}
			return templateParameters;
		}

		/// <summary>
		/// Parses the Documentation comment of code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element too parse the documentation comment for</param>
		/// <param name="linePrefix">The line prefix to use for the documentation comment</param>
		/// <returns>A <see cref="Simplabs.NetDocumentor.Documentation.DocumentationComment"/> that represents the documentation comment of code element <paramref name="element"/></returns>
		internal static DocumentationComment ParseDocumentationComment(CodeElement2 element, string linePrefix) {
			return BuildDocumentationComment(GetDocumentationCommentText(element), linePrefix);
		}

		/// <summary>
		/// Parses the Documentation comment of code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element too parse the documentation comment for</param>
		/// <param name="linePrefix">The line prefix to use for the documentation comment</param>
		/// <returns>A <see cref="Simplabs.NetDocumentor.Documentation.DocumentationComment"/> that represents the documentation comment of code element <paramref name="element"/></returns>
		internal static DocumentationComment ParseDocumentationComment(VCCodeElement element, string linePrefix) {
			if (element.InfoLocation == vsCMInfoLocation.vsCMInfoLocationExternal) {
				return BuildDocumentationComment(GetDocumentationCommentTextFromExternal(element), linePrefix);
			}
			bool hasDocumentationCommentAtDefinition = HasDocumentationComment(element, vsCMWhere.vsCMWhereDefinition);
			bool hasDocumentationCommentAtDeclaration = HasDocumentationComment(element, vsCMWhere.vsCMWhereDeclaration);
			if (!hasDocumentationCommentAtDeclaration && !hasDocumentationCommentAtDefinition) {
				return new DocumentationComment(linePrefix);
			}
			if (hasDocumentationCommentAtDeclaration) {
				return BuildDocumentationComment(GetDocumentationCommentText(element, vsCMWhere.vsCMWhereDeclaration), linePrefix);
			} else {
				return BuildDocumentationComment(GetDocumentationCommentText(element, vsCMWhere.vsCMWhereDefinition), linePrefix);
			}
		}

		/// <summary>
		/// Removes any existing comments from code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The element to remove existing comments from</param>
		internal static void RemoveExistingComment(CodeElement2 element) {
			EditPoint elementStartPoint = element.StartPoint.CreateEditPoint();
			EditPoint sourceCodePoint = elementStartPoint.CreateEditPoint();
			sourceCodePoint.LineUp(1);
			sourceCodePoint.StartOfLine();
			elementStartPoint.LineUp(1);
			elementStartPoint.EndOfLine();
			string lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
			switch (element.Language) {
				case CodeModelLanguageConstants.vsCMLanguageCSharp:
					if (lineText.StartsWith("//")) {
						while (lineText.StartsWith("//")) {
							sourceCodePoint.LineUp(1);
							lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
						}
						sourceCodePoint.EndOfLine();
						elementStartPoint.Delete(sourceCodePoint);
						return;
					} else if (lineText.EndsWith("*/")) {
						while (!lineText.Contains("/*")) {
							sourceCodePoint.LineUp(1);
							lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength);
						}
						sourceCodePoint.MoveToLineAndOffset(sourceCodePoint.Line, lineText.LastIndexOf("/*") + 1);
						elementStartPoint.Delete(sourceCodePoint);
					}
					break;
				case CodeModelLanguageConstants.vsCMLanguageVB:
					if (lineText.StartsWith("'")) {
						while (lineText.StartsWith("'")) {
							sourceCodePoint.LineUp(1);
							lineText = sourceCodePoint.GetText(sourceCodePoint.LineLength).Trim();
						}
						sourceCodePoint.EndOfLine();
						elementStartPoint.Delete(sourceCodePoint);
					}
					break;
				default:
					break;
			}
		}

		/// <summary>
		/// Throws an exception signalling that there were erros documenting code element <paramref name="element"/>
		/// </summary>
		/// <param name="element">The code element during thats documentation the error occured</param>
		/// <param name="exception">The exception that occured</param>
		internal static DocumentationException PrepareDocumentationException(object element, Exception exception) {
			CodeElement theElement = element as CodeElement;
			string exceptionText = string.Empty;
			if (element != null) {
				try {
					exceptionText = string.Format(
						CultureInfo.CurrentCulture,
						global::Simplabs.NetDocumentor.Properties.Resources.ErrorOccuredDuringDocumentation,
						theElement.FullName
					);
				} catch (Exception) {
					exceptionText = string.Format(
						CultureInfo.CurrentCulture,
						global::Simplabs.NetDocumentor.Properties.Resources.ErrorOccuredDuringDocumentation,
						"<unknown>"
					);
				}
			} else {
				exceptionText = string.Format(
					CultureInfo.CurrentCulture,
					global::Simplabs.NetDocumentor.Properties.Resources.ErrorOccuredDuringDocumentation,
					"<unknown>"
				);
			}
			return new DocumentationException(exceptionText, exception);
		}

		/// <summary>
		/// Finds a code element in a project item
		/// </summary>
		/// <param name="element">The code element to find</param>
		/// <param name="projectItem">The project item to find the code element in</param>
		/// <returns>The code element that has been found in the project item or null if no code element was found</returns>
		/// <remarks>This method is helpful when a code element is defined in more than one project items.</remarks>
		internal static VCCodeElement FindCodeElement(VCCodeElement element, ProjectItem projectItem) {
			VCCodeElement theElement;
			FindElement(element, (VCCodeElements)projectItem.FileCodeModel.CodeElements, out theElement);
			return theElement;
		}

		/// <summary>
		/// Finds a code element in a collection of elements
		/// </summary>
		/// <param name="codeElement">The code element to find</param>
		/// <param name="elementCollection">The collection of elements to find code element <paramref name="element"/> in</param>
		/// <returns>The element or null if the element could not be found</returns>
		private static void FindElement(VCCodeElement codeElement, VCCodeElements elementCollection, out VCCodeElement foundElement) {
			foreach (VCCodeElement element in elementCollection) {
				if (element.FullName == codeElement.FullName) {
					foundElement = element;
					return;
				}
				FindElement(codeElement, (VCCodeElements)element.Children, out foundElement);
			}
			foundElement = null;
		}

	}

}