using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell;
using EnvDTE;
using System.Diagnostics;

namespace Simplabs.NetDocumentor.IdeIntegration {

	/// <summary>
	/// Custom Task Provider of the Addin
	/// </summary>
	internal class TaskProvider : IVsTaskProvider {

		/// <summary>
		/// The Task list service of the IDE
		/// </summary>
		protected IVsTaskList taskList;

		/// <summary>
		/// The cookies that identified this task provider
		/// </summary>
		private uint taskListProviderCookie;

		/// <summary>
		/// The tasks the provider provides
		/// </summary>
		private List<Task> theTasks;

		/// <summary>
		/// Initializes a new TaskProvider
		/// </summary>
		internal TaskProvider() {
			taskList = ServiceProvider.GetService<IVsTaskList>();
			theTasks = new List<Task>();
			taskList.RegisterTaskProvider(this, out taskListProviderCookie);
		}

		/// <summary>
		/// Disposes of the elements of the task provider
		/// </summary>
		protected void Dispose() {
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// Disposes of the elements of the task provider
		/// </summary>
		/// <param name="disposing">Specifies whether the method shall actually deispose or not</param>
		protected virtual void Dispose(bool disposing) {
			try {
				if (disposing) {
					taskList.UnregisterTaskProvider(taskListProviderCookie);
				}
			} catch (Exception) {
			}
		}

		/// <summary>
		/// Standard enumerator over tasks supplied by this provider.
		/// </summary>
		/// <param name="ppenum">The task enumeration</param>
		/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
		int IVsTaskProvider.EnumTaskItems(out IVsEnumTaskItems ppenum) {
			ppenum = new TaskEnumerator(theTasks);
			return VSConstants.S_OK;
		}

		/// <summary>
		/// Returns the image list for the tasks of this task provider.
		/// </summary>
		/// <param name="phImageList">Pointer to hImageList, a handle for this image list.</param>
		/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
		int IVsTaskProvider.ImageList(out IntPtr phImageList) {
			phImageList = IntPtr.Zero;
			return VSConstants.S_OK;
		}

		/// <summary>
		/// Notifies the client to unregister the task provider from the task list if it has not been unregistered by the time the task list is closing.
		/// </summary>
		/// <param name="pTaskList">Pointer to the IVsTaskList interface that the task provider is registered to.</param>
		/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
		int IVsTaskProvider.OnTaskListFinalRelease(IVsTaskList pTaskList) {
			Dispose(true);
			return VSConstants.S_OK;
		}

		/// <summary>
		/// Not implemented
		/// </summary>
		/// <param name="pbstrKey"></param>
		/// <returns>VSConstants.S_FALSE</returns>
		int IVsTaskProvider.ReRegistrationKey(out string pbstrKey) {
			pbstrKey = this.GetType().FullName;
			return VSConstants.S_FALSE;
		}

		/// <summary>
		/// Returns the list of subcategories for the tasks of this task provider.
		/// </summary>
		/// <param name="cbstr">Number of subcategories in the list.</param>
		/// <param name="rgbstr">Array of BSTRs whose elements represent the subcategories indicated by cbstr.</param>
		/// <param name="pcActual">Pointer to the number of BSTRs contained in the rgbstrarray.</param>
		/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
		int IVsTaskProvider.SubcategoryList(uint cbstr, string[] rgbstr, out uint pcActual) {
			pcActual = 0;
			return VSConstants.S_OK;
		}

		/// <summary>
		/// Adds a task to the task list
		/// </summary>
		/// <param name="item">The project item to add the task for</param>
		protected void AddTask(Task theTask) {
			theTasks.Add(theTask);
		}

		/// <summary>
		/// Flushes and refreshes all Tasks
		/// </summary>
		protected void FlushAndRefresh() {
			taskList.RefreshTasks(taskListProviderCookie);
		}

		/// <summary>
		/// A task
		/// </summary>
		internal class Task : IVsTaskItem {

			/// <summary>
			/// The text of the task
			/// </summary>
			private string taskText;

			/// <summary>
			/// The line the task refers to
			/// </summary>
			private int taskLine;

			/// <summary>
			/// The priority of the task
			/// </summary>
			private TaskPriority taskPriority;

			/// <summary>
			/// The project item the task refers to
			/// </summary>
			protected ProjectItem taskProjectItem;

			/// <summary>
			/// The category of the task
			/// </summary>
			private VSTASKCATEGORY taskCategory;

			/// <summary>
			/// Holds whether the task is checked in the task list or nor
			/// </summary>
			private bool taskIsChecked;

			/// <summary>
			/// Initializes a new task
			/// </summary>
			/// <param name="projectItem">The project item the task refers to</param>
			/// <param name="text">The text of the task</param>
			/// <param name="line">The line the task refers to</param>
			/// <param name="priority">The priority of the task</param>
			/// <param name="category">The category of the task</param>
			internal Task(ProjectItem projectItem, string text, int line, TaskPriority priority, VSTASKCATEGORY category) {
				taskText = text;
				taskProjectItem = projectItem;
				taskLine = line;
				taskPriority = priority;
				taskCategory = category;
			}
				
			/// <summary>
			/// Gets whether the task can be deleted
			/// </summary>
			/// <param name="pfCanDelete">Is set to 1 if the task can be deleted or otherwise to 0</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.CanDelete(out int pfCanDelete) {
	 			pfCanDelete = 1;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the category of the task
			/// </summary>
			/// <param name="pCat">Is set to the category of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.Category(VSTASKCATEGORY[] pCat) {
				pCat[0] = taskCategory;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the column of the task
			/// </summary>
			/// <param name="piCol">Is set to the column of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.Column(out int piCol) {
				piCol = -1;
				return VSConstants.S_FALSE;
			}

			/// <summary>
			/// Gets the file name of the task
			/// </summary>
			/// <param name="pbstrMkDocument">Is set to the file name of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.Document(out string pbstrMkDocument) {
				pbstrMkDocument = taskProjectItem.Name;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets whether the task has help
			/// </summary>
			/// <param name="pfHasHelp">Is set to 1 if the task has help or 0 otherwise</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.HasHelp(out int pfHasHelp) {
				pfHasHelp = 0;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the task's image's index in the image list
			/// </summary>
			/// <param name="pIndex">Is set to the image's index in the image list</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.ImageListIndex(out int pIndex) {
				pIndex = -1;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets whether the task is readonly
			/// </summary>
			/// <param name="field">The field that is being checked for whether it is readonly</param>
			/// <param name="pfReadOnly">Is set to 1 if the task is readonly or 0 otherwise</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.IsReadOnly(VSTASKFIELD field, out int pfReadOnly) {
				pfReadOnly = 1;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the line the task refers to
			/// </summary>
			/// <param name="piLine">Is set to the line the task refers to</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			public virtual int Line(out int piLine) {
				piLine = taskLine;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Navigates to the item that is associated with the task
			/// </summary>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			public virtual int NavigateTo() {
				return VSConstants.S_FALSE;
			}

			/// <summary>
			/// Navigates to the help topic associated with the task
			/// </summary>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.NavigateToHelp() {
				return VSConstants.S_FALSE;
			}

			/// <summary>
			/// Deletes the task
			/// </summary>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.OnDeleteTask() {
				return VSConstants.S_FALSE;
			}

			/// <summary>
			/// Filters the tasks in the list
			/// </summary>
			/// <param name="fVisible"></param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.OnFilterTask(int fVisible) {
				return VSConstants.S_FALSE;
			}

			/// <summary>
			/// Gets the subcategory index of the task
			/// </summary>
			/// <param name="pIndex">The subcategory index</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.SubcategoryIndex(out int pIndex) {
				pIndex = -1;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets whether the task is checked
			/// </summary>
			/// <param name="pfChecked">True if the task is checked, false otherwise</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.get_Checked(out int pfChecked) {
				pfChecked = taskIsChecked ? 1 : 0;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the priority of the task
			/// </summary>
			/// <param name="ptpPriority">The priority of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.get_Priority(VSTASKPRIORITY[] ptpPriority) {
				if (ptpPriority != null && ptpPriority.Length > 0) {
					ptpPriority[0] = (VSTASKPRIORITY)taskPriority;
				}
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the taxt of the task
			/// </summary>
			/// <param name="pbstrName">The text of the tsk</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.get_Text(out string pbstrName) {
				pbstrName = taskText;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Sets whether the task is checked
			/// </summary>
			/// <param name="fChecked">1 if the task should be checked, false otherwise</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.put_Checked(int fChecked) {
				taskIsChecked = fChecked == 1 ? true : false;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Sets the priority of the task
			/// </summary>
			/// <param name="tpPriority">The priority of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsTaskItem.put_Priority(VSTASKPRIORITY tpPriority) {
				taskPriority = (TaskPriority)tpPriority;
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Sets the text of the task
			/// </summary>
			/// <param name="bstrName">The text of the task</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			public int put_Text(string bstrName) {
				taskText = bstrName;
				return VSConstants.S_OK;
			}

		}

		/// <summary>
		/// An enumerator that iterates over a collection of tasks
		/// </summary>
		internal class TaskEnumerator : IVsEnumTaskItems {

			/// <summary>
			/// The tasks the enumerator enumerates over
			/// </summary>
			IEnumerable<Task> theTasks;

			/// <summary>
			/// The actual enumerator
			/// </summary>
			IEnumerator<Task> enumerator;

			/// <summary>
			/// Initializes a new TaskEnumerator
			/// </summary>
			/// <param name="tasks">The tasks the enumerator enumerates over</param>
			internal TaskEnumerator(IEnumerable<Task> tasks) {
				theTasks = tasks;
				enumerator = theTasks.GetEnumerator();
			}

			/// <summary>
			/// Clones the enumerator
			/// </summary>
			/// <param name="ppenum">The cloned enumerator</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsEnumTaskItems.Clone(out IVsEnumTaskItems ppenum) {
				ppenum = new TaskEnumerator(theTasks);
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Gets the next <paramref name="celt"/> elements from the enumerator
			/// </summary>
			/// <param name="celt">The number of elements to get</param>
			/// <param name="rgelt">The elements</param>
			/// <param name="pceltFetched">The number of elements written to <paramref name="rgelt"/></param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsEnumTaskItems.Next(uint celt, IVsTaskItem[] rgelt, uint[] pceltFetched) {
				if (rgelt == null || rgelt.Length < celt) {
					return VSConstants.S_FALSE;
				}
				uint fetched = 0;
				while (fetched < celt && enumerator.MoveNext()) {
					rgelt[fetched++] = enumerator.Current;
				}
				if (pceltFetched != null && pceltFetched.Length > 0) {
					pceltFetched[0] = fetched;
				}
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Resets the enumerator
			/// </summary>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsEnumTaskItems.Reset() {
				enumerator.Reset();
				return VSConstants.S_OK;
			}

			/// <summary>
			/// Skips <paramref name="celt"/> elements
			/// </summary>
			/// <param name="celt">The number of elements to skip</param>
			/// <returns>VSConstants.S_OK on success, VSConstants.S_FAIL otherwise</returns>
			int IVsEnumTaskItems.Skip(uint celt) {
				while (celt-- >= 0) {
					if (!enumerator.MoveNext() && celt != 0) {
						return VSConstants.S_FALSE;
					}
				}
				return VSConstants.S_OK;
			}

		}

	}

}