using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using EnvDTE;
using System.Globalization;
using Simplabs.NetDocumentor.Documentation;

namespace Simplabs.NetDocumentor.UI {

	/// <summary>
	/// Form that displays options and progress for the completion of documentation for all code elements in a project
	/// </summary>
	internal partial class CompletionRunForm : Form {

		/// <summary>
		/// Holds whether the process is currently running
		/// </summary>
		private bool isRunning;

		/// <summary>
		/// The runner that actually performs the work
		/// </summary>
		private CompletionRunner theRunner;

		/// <summary>
		/// The thread <see cref="theRunner"/> is running in
		/// </summary>
		private BackgroundWorker workerThread;

		/// <summary>
		/// Initializes a new CompletionRunForm
		/// </summary>
		/// <param name="theProject">The project for that's code elements documentation is to be completed</param>
		/// <param name="runner">The runner that actually performs the work</param>
		internal CompletionRunForm(Project theProject, CompletionRunner runner) {
			InitializeComponent();
			Text = string.Format(
				CultureInfo.CurrentUICulture,
				global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormCaption,
				theProject.Name
			);
			okButton.Text = global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormOkButtonCaption;
			cancelButton.Text = global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormCancelButtonCaption;
			this.cancelButton.Click += new System.EventHandler(this.cancelButton_Click_Cancel);
			insertTaskShortcutsLabel.Text = global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormInsertTaskSortcutsLabelCaption;
			insertTaskShortcutsNoRadioButton.Text = global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormInsertTaskSortcutsNoCaption;
			insertTaskShortcutsYesRadioButton.Text = global::Simplabs.NetDocumentor.Properties.Resources.CompletionRunFormInsertTaskSortcutsYesCaption;
			theRunner = runner;
			progressBar.Maximum = theRunner.NumberOfItemsToDocument;
		}

		/// <summary>
		/// Determines whether the specified key is a regular input key or a special key that requires preprocessing.
		/// </summary>
		/// <param name="keyData">One of the <see cref="System.Windows.Forms.Keys"/> values.</param>
		/// <returns>True if the specified key is a regular input key; otherwise, false.</returns>
		protected override bool IsInputKey(Keys keyData) {
			if (keyData == Keys.Escape) {
				return true;
			}
			return base.IsInputKey(keyData);
		}

		/// <summary>
		/// Raises the KeyDown event.
		/// </summary>
		/// <param name="e">A KeyEventArgs that contains the event data.</param>
		protected override void OnKeyDown(KeyEventArgs e) {
			if (!isRunning && e.KeyCode == Keys.Escape) {
				DialogResult = DialogResult.Cancel;
				Close();
			}
			base.OnKeyDown(e);
		}

		/// <summary>
		/// Handles the Click event of the "ok" button
		/// </summary>
		/// <param name="sender">The sender of the event</param>
		/// <param name="e">Arguments of the event</param>
		private void okButton_Click(object sender, EventArgs e) {
			workerThread = new BackgroundWorker();
			workerThread.WorkerSupportsCancellation = true;
			workerThread.DoWork += delegate {
				theRunner.Run(insertTaskShortcutsYesRadioButton.Checked, workerThread);
			};
			workerThread.ProgressChanged += delegate {
				progressBar.PerformStep();
			};
			isRunning = true;
			okButton.Enabled = false;
			insertTaskShortcutsNoRadioButton.Enabled = false;
			insertTaskShortcutsYesRadioButton.Enabled = false;
			insertTaskShortcutsLabel.Enabled = false;
			workerThread.RunWorkerCompleted += delegate {
				DialogResult = DialogResult.OK;
				Close();
			};
			workerThread.RunWorkerAsync();
		}

		/// <summary>
		/// Handles the Click event of the "cancel" button
		/// </summary>
		/// <param name="sender">The sender of the event</param>
		/// <param name="e">Arguments of the event</param>
		private void cancelButton_Click_Cancel(object sender, EventArgs e) {
			if (isRunning) {
				workerThread.CancelAsync();
			}
			DialogResult = DialogResult.Cancel;
			Close();
		}

	}

}