using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Simplabs.NetDocumentor.Settings;

namespace Simplabs.NetDocumentor.UI {

	/// <summary>
	/// The base options page for the AddIn
	/// </summary>
	public abstract partial class OptionsPage : UserControl, EnvDTE.IDTToolsOptionsPage {

		/// <summary>
		/// The features that can be controlled thru the options page
		/// </summary>
		private List<Feature> features;

		/// <summary>
		/// Initializes a new OptionsPage
		/// </summary>
		protected OptionsPage() {
			InitializeComponent();
			labelUnavailable.Text = global::Simplabs.NetDocumentor.Properties.Resources.OptionsPageAddinUnavailable;
			groupBoxFeatureDescription.Text = global::Simplabs.NetDocumentor.Properties.Resources.OptionsPageFeatureDescriptionCaption;
			labelCaption.Text = global::Simplabs.NetDocumentor.Properties.Resources.OptionsPageCaption;
			base.Paint += new PaintEventHandler(delegate {
				treeViewOptions.Refresh();
			});
		}

		/// <summary>
		/// Gets or sets the features that can be controlled thru the options page
		/// </summary>
		internal List<Feature> Features {
			get {
				return features;
			}
			set {
				features = value;
			}
		}

		/// <summary>
		/// Performs manual initialization of the form
		/// </summary>
		protected virtual void ManualInitialize() {
		}

		/// <summary>
		/// Saves custom options
		/// </summary>
		protected virtual void SaveCustomOptions() {
		}

		/// <summary>
		/// Initializes the tree view (sets up the nodes with appropriate caption and state)
		/// </summary>
		private void InitializeTreeView() {
			treeViewOptions.CheckBoxes = true;
			treeViewOptions.Nodes.Clear();
			foreach (Feature feature in features) {
				TreeNode node = new TreeNode(feature.Name);
				node.Tag = feature;
				node.Checked = feature.IsEnabled;
				treeViewOptions.Nodes.Add(node);
			}
		}

		/// <summary>
		/// Returns an object containing all properties contained on the specified custom Tools Options page.
		/// </summary>
		/// <param name="PropertiesObject">The Tools Options object.</param>
		void EnvDTE.IDTToolsOptionsPage.GetProperties(ref object PropertiesObject) {
		}

		/// <summary>
		/// Handles the AfterCreated event of the options page
		/// </summary>
		/// <param name="DTEObject">The parent object of the Tools Options page, namely, a EnvDTE.DTE object.</param>
		void EnvDTE.IDTToolsOptionsPage.OnAfterCreated(EnvDTE.DTE DTEObject) {
			ManualInitialize();
			if (!Addin.IsConnected) {
				panelUnavailable.Visible = true;
			} else {
				panelUnavailable.Visible = false;
				InitializeTreeView();
				treeViewOptions.AfterSelect += delegate(object sender, TreeViewEventArgs e) {
					labelFeatureDescription.Text = ((Feature)e.Node.Tag).Description;
				};
				treeViewOptions.AfterCheck += delegate(object sender, TreeViewEventArgs e) {
					((Feature)e.Node.Tag).IsEnabled = e.Node.Checked;
				};
			}
		}

		/// <summary>
		/// Handles the Cancel event of the options page
		/// </summary>
		void EnvDTE.IDTToolsOptionsPage.OnCancel() {
			global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.Reload();
		}

		/// <summary>
		/// Handles the Help event of the options page
		/// </summary>
		void EnvDTE.IDTToolsOptionsPage.OnHelp() {
		}

		/// <summary>
		/// Handles the OK event of the options page
		/// </summary>
		void EnvDTE.IDTToolsOptionsPage.OnOK() {
			SaveCustomOptions();
			global::Simplabs.NetDocumentor.Properties.FeatureSettings.Default.Save();
		}

	}

}