//---------------------------------------------------------------------
//  This file is part of the Microsoft .NET Framework SDK Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
// 
//This source code is intended only as a supplement to Microsoft
//Development Tools and/or on-line documentation.  See these other
//materials for detailed information regarding Microsoft code samples.
// 
//THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
//KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//PARTICULAR PURPOSE.
//---------------------------------------------------------------------

using System;
using System.Text;

namespace Microsoft.VisualStudio.Samples.TextEncoding.EncodingApplication
{
    /// <summary>
    /// Best fit fallback defining custom encoding
    /// </summary>
    /// <seealso cref="T:System.Text.EncoderFallback"/>
	public class EncoderBestFitFallback :EncoderFallback
	{

        /// <summary>
        /// </summary>
		private Encoding parentEncoding=null;

        /// <summary>
        /// Initialisiert eine neue Microsoft.VisualStudio.Samples.TextEncoding.EncodingApplication.EncoderBestFitFallback
        /// </summary>
        /// <param name="targetEncoding"></param>
		public EncoderBestFitFallback(Encoding targetEncoding)
		{
			this.ParentEncoding = targetEncoding;
		}

        /// <summary>
        /// Liest oder setzt die
        /// </summary>
        /// <value></value>
		public Encoding ParentEncoding
		{
			set
			{
				this.parentEncoding = (Encoding)value.Clone();
				this.parentEncoding.EncoderFallback=new EncoderReplacementFallback("");
				this.parentEncoding.DecoderFallback=new DecoderReplacementFallback("");
			}

			get
			{
				return this.parentEncoding;
			}            
		}

        /// <summary>
        /// Liest die
        /// </summary>
        /// <value></value>
		public override int MaxCharCount
		{
			get
			{
				return 18;
			}
		}

        /// <summary>
        /// When overridden in a derived class, initializes a new instance of the <see cref="T:System.Text.EncoderFallbackBuffer"></see> class.
        /// </summary>
        /// <returns>An <see cref="T:System.Text.EncoderFallbackBuffer"></see> object.</returns>
        /// <filterpriority>2</filterpriority>
		public override EncoderFallbackBuffer CreateFallbackBuffer()
		{
			return new EncoderBestFitFallbackBuffer(this);
		}
	}

    /// <summary>
    /// Class to encapsulate the buffer for best fit fallback
    /// </summary>
    /// <seealso cref="T:System.Text.EncoderFallbackBuffer"/>
	public sealed class EncoderBestFitFallbackBuffer: EncoderFallbackBuffer
	{
        /// <summary>
        /// </summary>
		private String strFallback = String.Empty;
        /// <summary>
        /// </summary>
		private int charIndex = 1;
        /// <summary>
        /// </summary>
		private Encoding parentEncoding;

        /// <summary>
        /// Initialisiert eine neue Microsoft.VisualStudio.Samples.TextEncoding.EncodingApplication.EncoderBestFitFallbackBuffer
        /// </summary>
        /// <param name="fallback"></param>
		public EncoderBestFitFallbackBuffer(EncoderBestFitFallback fallback)
		{
			parentEncoding = fallback.ParentEncoding;
		}

        /// <summary>
        /// When overridden in a derived class, prepares the fallback buffer to handle the specified input character.
        /// </summary>
        /// <param name="charUnknown">An input character.</param>
        /// <param name="index">The index position of the character in the input buffer.</param>
        /// <returns>true if the fallback buffer can process charUnknown; false if the fallback buffer ignores charUnknown.</returns>
        /// <filterpriority>1</filterpriority>
		public override bool Fallback(char charUnknown, int index)
		{
			// Since both fallback methods require normalizing a string, make a string out of our char
			String strUnknown = new String(charUnknown, 1);

			return Fallback(strUnknown);
		}

        /// <summary>
        /// When overridden in a derived class, prepares the fallback buffer to handle the specified surrogate pair.
        /// </summary>
        /// <param name="charUnknownHigh">The high surrogate of the input pair.</param>
        /// <param name="charUnknownLow">The low surrogate of the input pair.</param>
        /// <param name="index">The index position of the surrogate pair in the input buffer.</param>
        /// <returns>true if the fallback buffer can process charUnknownHigh and charUnknownLow; false if the fallback buffer ignores the surrogate pair.</returns>
        /// <filterpriority>1</filterpriority>
		public override bool Fallback(char charUnknownHigh, char charUnknownLow, int index)
		{
			// Since both fallback methods require normalizing a string, make a string out of our chars
			String strUnknown = new String(new char[] { charUnknownHigh, charUnknownLow });

			return Fallback(strUnknown);
		}

        /// <summary>
        /// </summary>
        /// <param name="strUnknown"></param>
        /// <returns></returns>
		private bool Fallback(String strUnknown)
		{
			// If they were falling back already they need to fail
			if (charIndex <= strFallback.Length)
			{
				// Throw it, using our complete character
				charIndex = 1;
				strFallback = String.Empty;
				throw new ArgumentException("Unexpected recursive fallback", "chars");
			}

			// They need to fallback our character
			String strNormal = "";
			try
			{
				// Normalize our character
				strNormal = strUnknown.Normalize(NormalizationForm.FormKD);

				// If it didn't change we'll still have to be ?
				if (strNormal == strUnknown)
				{
					strNormal = "";
				}
			}
				// Illegal Unicode can throw a normalization exception
			catch (ArgumentException)
			{
				// Allow the string to become a ? fallback
			}

			// Try encoding the normalized string
			strFallback = parentEncoding.GetString(parentEncoding.GetBytes(strNormal));

			// Any data from this fallback?
			if ((strFallback.Length == 0) || (strFallback[0] != strNormal[0]))
			{
				// We didn't have data or didn't have the first character, so use ?
				strFallback = "?";
			}

			charIndex = 0;

			return true;
		}

        /// <summary>
        /// When overridden in a derived class, retrieves the next character in the fallback buffer.
        /// </summary>
        /// <returns>The next character in the fallback buffer.</returns>
        /// <filterpriority>2</filterpriority>
		public override char GetNextChar()
		{
			// If we're above string length then return nothing
			if (charIndex >= strFallback.Length)
			{
				// We use charIndex = length + 1 as a flag to tell us we're long past done.
				if (charIndex == strFallback.Length)
					charIndex++;
				return '\0';
			}

			// Return the character at our charIndex (& increment it)
			return strFallback[charIndex++];
		}

        /// <summary>
        /// When overridden in a derived class, causes the next call to the <see cref="M:System.Text.EncoderFallbackBuffer.GetNextChar"></see> method to access the data buffer character position that is prior to the current character position.
        /// </summary>
        /// <returns>true if the <see cref="M:System.Text.EncoderFallbackBuffer.MovePrevious"></see> operation was successful; otherwise, false.</returns>
        /// <filterpriority>1</filterpriority>
		public override bool MovePrevious()
		{
			// Back up only if we're within the string or just did the last char
			if (charIndex <= strFallback.Length)
				charIndex--;

			// Return true if we were able to move previous
			return (charIndex >= 0 || charIndex < strFallback.Length);
		}


        /// <summary>
        /// Liest die
        /// </summary>
        /// <value></value>
		public override int Remaining
		{
			get
			{
				// Any characters remaining?  If so, how many?
				if (charIndex < strFallback.Length)
					return strFallback.Length - charIndex;

				return 0;
			}
		}

        /// <summary>
        /// Initializes all data and state information pertaining to this fallback buffer.
        /// </summary>
        /// <filterpriority>1</filterpriority>
		public override void Reset()
		{
			strFallback = String.Empty;
			charIndex = 1;
		}

	}
}
